# ===================================================================================
#                            animationutilities.py                                  |
# ===================================================================================
#                                                                                   |
# Programmer:     Sandy Wilson (KnightErrant)                                       |
#                                                                                   |
# Creation Date:  18 May  2007                                                      |
# Revision Dates: Many revisions since creation.                                    |
#                 05 June 2007 - Finished converttxttocas function.                 |
#                                                                                   |
#                                Removed the 'append' opening for convertcastottxt  |
#                                so it will overwrite.  The other style was for     |
#                                doing directory summaries when the output txt file |
#                                is opened and closed many times.                   |
#                                                                                   |
#                                Finished the animmerge updates so that it will     |
#                                now reorder data coming from a cas file when the   |
#                                bone order conflicts with the .ms3d file.  This    |
#                                is because we want to deal with zxiang1983's use   |
#                                of animations from new_2h_axe_halberd which have   |
#                                the jaw and eyebrow bones switched.                |
#                                                                                   |
#                 15 June 2007 - Added a reorder bones utility to move new bones    |
#                                in front of the weapon/shield bones.               |
#                                (Tested this, it works with a 2 bone tail for the  |
#                                armored_sergeants_lod0.mesh.)  Had to modify the   |
#                                meshconverter to Version 1_0_4 to do new node      |
#                                numbers for the moved weapon/shield bones.         |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
#    A collection of animation utilities selectable from a simple GUI interface.    |
#                                                                                   |
# ===================================================================================

# Imports.
import os
import os.path
import array
import struct                                    # This is for reading binary files.
from Tkinter import *                            # This is for the main GUI.
from animationlibrary import *                   # Accesses the library functions. 
from tkSimpleDialog import Dialog                # Used for simple dialog GUIs.
from tkSimpleDialog import askinteger            # Used for simple dialog GUIs to get an integer.
from tkMessageBox   import showinfo              # Used for the info button help dialogs. 
from tkMessageBox   import askokcancel           # Used for proceed or exit in reorderbones.

# ===================================================================================
#                            Applications.                                          |
# ===================================================================================
#                                                                                   |
# animmerge() - Merges an animation with a Milkshape .ms3d file.  Users can then    |
#               modify the animation within Milkshape.  The .ms3d and .cas files    |
#               are selected from GUI file choosers and then a dialog box asks for  |
#               the animation Frames Per Second desired.  20 is at game speed while |
#               smaller values are better for studying the animation.               |
#                                                                                   |
#               NOTE: The header data is stored as an ASCII string of tokens in     |
#                     joint comment[0].                                             |
#                     The hierarchy tree data is stored in joint comment[1].        |
#                     The footer data, if present, is stored in joint comment[2].   |
#                                                                                   |
# Functions: makeheaderdatastring( headerdata )                                     |
#            makefooterdatastring( footerdata )                                     |
#            makehierarchydatastring( hierarchydata )                               |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
# animextract() - Takes animation data in an .ms3d file and reoutputs it as a .cas  |
#                 file.  It reads the header, hierarchy tree, and footer from the   |
#                 joint comments [0], [1], and [2], respectively.  (No footer if    |
#                 there is no joint comment[2].)  The anim time in the header is    |
#                 recomputed in case the user has added frames.                     |
#                                                                                   |
# Functions: getcasheaderdata( signaturebyte1, signaturebyte2, animtime )           |
#            getcashierarchytree()                                                  |
#            getcastimeticks( nframes )                                             |
#            getcasbonesection( jointdata )                                         |
#            getcasdata( jointdata )                                                |
#            getcasfooter()                                                         |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
# exportskeleton() - Exports a skeleton stored in an ASCII .skeleton file to all    |
#                    the .cas files in a user GUI-selected directory.  It also      |
#                    puts up a dialog box for the user to select scaling factors    |
#                    for the x, y, and z animation data.  This is so the user can   |
#                    shorten the y-values for a dwarf, for instance, so his feet    |
#                    will touch the ground.                                         |
#                                                                                   |
# Functions: Dialog( Toplevel )                                                     |
#            getscalingfactors( Dialog )                                            |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
# extractskeleton() - Extracts the bones from an .ms3d file and stores the x, y,    |
#                     and z values and the bone name in the file                    |
#                     'extracted.skeleton'.  The user is expected to rename this    |
#                     to something meaningful based on the unit he is extracting    |
#                     from.  Note that the skeleton has the x-values with opposite  |
#                     signs to what is in the .ms3d file.  This is because the      |
#                     skeleton is meant to be exported to CA .cas files with the    |
#                     function exportskeleton() so the convention is chosen to      |
#                     match CA's convention.                                        |
#                                                                                   |
# Functions: gethumanhierarchynumber( bonename )                                    |
#            getcamelhierarchynumber( bonename )                                    |
#            gethorsehierarchynumber( bonename )                                    |
#            getelephanthierarchynumber( bonename )                                 |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
# convertcastotxt() - Converts a binary .cas file to an ASCII equivalant text       |
#                     representation.                                               |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
# surveycasdirectory() - Reads all the .cas files in a user GUI-selected            |
#                        directory and creates a single .txt file containing        |
#                        a summary of the data based on a dialog where the          |
#                        user can select what data he wants: header,                |
#                        hierarchytree, time ticks, bone section, all data,         |
#                        and footers.  For surveying usually only the headers       |
#                        and footers are selected.                                  |
#                                                                                   |
# Functions: dataflags()                                                            |
#            getdataoptions( Dialog )                                               |
#            isfileindescrskeleton( fncas )                                         |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
# converttxttocas() - Converts a .txt representation of a .cas file back to its     |
#                     binary form.  Useful for making quick checks of the effects   |
#                     of changing a bone or two or some anim data.  Not intended    |
#                     for real animation work; that should be done with Milkshape   |
#                     and the modified animation extracted with animextract().      |
#                                                                                   |
# Functions: gettxtheaderdata( headerdatastring )                                   |
#            gettxthierarchytree( hierarchydatastring )                             |
#            gettxttimeticksdata( timeticksdatastring )                             |
#            gettxtbonesection( fidtxt, nbones )                                    |
#            geteulerdata( fidtxt, bonedata )                                       |
#            getanimdata( fidtxt, bonedata )                                        |
#            getposedata( fidtxt, bonedata )                                        |
#            gettxtfooterdata( footerdatastring )                                   |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
# reorderbones() - Reorders the bones in a .ms3d file.                              |
#                                                                                   |
# ----------------------------------------------------------------------------------|
#                                                                                   |
# exportextrabonesskeleton() - Exports a skeleton with extra bones.  This requires  |
#                              hierarchy data in the skeleton and new bone names    |
#                              hence the .skeleton file format of x, y, z offsets,  |
#                              hierarchy number, and bone name.                     |
#                                                                                   |
# ===================================================================================


# -----------------------------------------------------------------------------------
#                            makeheaderdatastring()                                 |
# -----------------------------------------------------------------------------------
def makeheaderdatastring( headerdata ) :

    # Extract the data.
    float_version          = headerdata[0]
    int_thirtyeight        = headerdata[1] 
    int_nine               = headerdata[2] 
    int_zero1              = headerdata[3] 
    float_animtime         = headerdata[4] 
    int_one1               = headerdata[5] 
    int_zero2              = headerdata[6] 
    signaturebytetriple1   = headerdata[7] 
    int_one2               = headerdata[8] 
    int_zero3              = headerdata[9] 
    signaturebytetriple2   = headerdata[10]

    # Write the fields as one string.
    string1                = formatfloat(float_version, 5, 3) + ' ' + str(int_thirtyeight).ljust(3) + ' ' + str(int_nine).ljust(3) + ' '
    string1                = string1 + str(int_zero1).ljust(3) + ' ' + formatfloat(float_animtime, 5, 3) + ' '
    string1                = string1 + str(int_one1).ljust(3) + ' ' + str(int_zero2).ljust(3) + '    ' 
    string1                = string1 + str(signaturebytetriple1[0]).ljust(3) + ' ' + str(signaturebytetriple1[1]).ljust(3) + ' ' + str(signaturebytetriple1[2]).ljust(3) + '    ' 
    string1                = string1 + str(int_one2).ljust(3) + ' ' + str(int_zero3).ljust(3) + '    '
    string1                = string1 + str(signaturebytetriple2[0]).ljust(3) + ' ' + str(signaturebytetriple2[1]).ljust(3) + ' ' + str(signaturebytetriple2[2]).ljust(3) + '\n' 

    return string1


# -----------------------------------------------------------------------------------
#                            makefooterdatastring()                                 |
# -----------------------------------------------------------------------------------
def makefooterdatastring( footerdata ) :

    # Extract the the first field to find out which kind of footer we have
    # to write: 
    #     (1) regular  - int_104 is anything but 18 or 170.
    #     (2) camel    - int_104 is actually 18.
    #     (3) fs_horse - int_104 is actually 170.
    int_104                = footerdata[0]

    if int_104 == 18 :
        # Camel case.
        int_one1           = footerdata[1]
        int_zero1          = footerdata[2]
        int_zero2          = footerdata[3]
        short_zero1        = footerdata[4]
        int_vec1           = footerdata[5]
        int_vec2           = footerdata[6]
        int_vec3           = footerdata[7]
        int_vec4           = footerdata[8]
        int_vec5           = footerdata[9]

        # Write the footer as one line for a cas to text conversion.
        string             = str(int_104).ljust(3) + ' ' + str(int_one1).ljust(2) + ' ' + str(int_zero1).ljust(2) + ' ' + str(int_zero2).ljust(2) + ' ' + str(short_zero1).ljust(2) + ' ' 

        for ii in range( len(int_vec1) ):
            string         = string + str(int_vec1[ii]).ljust(2) + ' ' 
        for ii in range( len(int_vec2) ):
            string         = string + str(int_vec2[ii]).ljust(2) + ' ' 
        for ii in range( len(int_vec3) ):
            string         = string + str(int_vec3[ii]).ljust(2) + ' ' 
        for ii in range( len(int_vec4) ):
            string         = string + str(int_vec4[ii]).ljust(2) + ' ' 
        for ii in range( len(int_vec5) ):
            string         = string + str(int_vec5[ii]).ljust(2) + '\n' 

        return string

    # Now it's either a regular anim or an fs_horse anim.  Pull out the common data
    # fields from the start of footerdata.
    int_one1               = footerdata[1]                 # Index 1.
    int_one2               = footerdata[2]                 # Index 2.
    AttribString           = footerdata[3]                 # Index 3. 
    int_one3               = footerdata[4]                 # Index 4.
    byte_zero1             = footerdata[5]                 # Index 5.
    int_one4               = footerdata[6]                 # Index 6.
    int_zero1              = footerdata[7]                 # Index 7.
    byte_zero2             = footerdata[8]                 # Index 8.
                                                         
    float_vec1             = footerdata[9]                 # Index 9.
                                                         
    int_zero2              = footerdata[10]                # Index 10.
    short_zero1            = footerdata[11]                # Index 11.
    int_minus1             = footerdata[12]                # Index 12.
    int_zero3              = footerdata[13]                # Index 13.

    # Start the string representation of the footer.
    string                 = str(int_104).ljust(2) + ' ' + str(int_one1).ljust(2) + ' ' + str(int_one2).ljust(2) + ' ' + AttribString + '   '                                                                 
    string                 = string + str(int_one3).ljust(2) + ' ' + str(byte_zero1).ljust(2) + ' ' + str(int_one4).ljust(2) + ' ' + str(int_zero1).ljust(2) + ' ' + str(byte_zero2).ljust(2) + ' '                  
    for ii in range( len(float_vec1) ) :                                                                                                                                                           
        string             = string + formatfloat( float_vec1[ii], 7, 5 ) + ' '                                                                                                                                       
                                                                                                                                                                                                       
    string                 = string + str(int_zero2).ljust(2) + ' ' + str(short_zero1).ljust(2) + ' '                                                                                                                 
    string                 = string + str(int_minus1).ljust(2) + ' ' + str(int_zero3).ljust(2) + ' '                                                                                                                 


    if int_104 == 170 :
        # It's an fs_horse type of anim.  Lots more data fields.
        string2            = footerdata[14]                # Index 14.
        int_one5           = footerdata[15]                # Index 15.
        byte_zero3         = footerdata[16]                # Index 16.
        int_one6           = footerdata[17]                # Index 17.
        int_zero4          = footerdata[18]                # Index 18.
        byte_zero4         = footerdata[19]                # Index 19.
        float_vec2         = footerdata[20]                # Index 20.
        int_vec1           = footerdata[21]                # Index 21.
        int_vec2           = footerdata[22]                # Index 22.
        int_vec3           = footerdata[23]                # Index 23.
        int_vec4           = footerdata[24]                # Index 24.
        int_vec5           = footerdata[25]                # Index 25.
        int_count          = footerdata[26]                # Index 26.
        int_five           = footerdata[27]                # Index 27.
        int_one7           = footerdata[28]                # Index 28.
        int_zero5          = footerdata[29]                # Index 29.
        float_vec3         = footerdata[30]                # Index 30.
        int_vec6           = footerdata[31]                # Index 31.
        byte_zero5         = footerdata[32]                # Index 32.
        int_vec7           = footerdata[33]                # Index 33.

        # Make the string representation of the fs_horse footer. 
        string             = string + ' ' + string2 + ' ' + str(int_one5).ljust(2) + ' ' + str(byte_zero3).ljust(2) + ' ' + str(int_one6).ljust(2) + ' ' + str(int_zero4).ljust(2) + ' ' + str(byte_zero4).ljust(2) + ' ' 
        
        for ii in range( len(float_vec2) ):
            string         = string + formatfloat(float_vec2[ii],7,5) + '  ' 

        for ii in range( len(int_vec1) ):
            string         = string + str(int_vec1[ii]).ljust(2) + ' ' 

        for ii in range( len(int_vec2) ):
            string         = string + str(int_vec2[ii]).ljust(2) + ' ' 

        for ii in range( len(int_vec3) ):
            string         = string + str(int_vec3[ii]).ljust(2) + ' ' 

        for ii in range( len(int_vec4) ):
            string         = string + str(int_vec4[ii]).ljust(2) + ' ' 

        for ii in range( len(int_vec5) ):
            string         = string + str(int_vec5[ii]).ljust(2) + ' ' 

        string             = string + str(int_count).ljust(2) + ' ' + str(int_five).ljust(2) + ' ' + str(int_one7).ljust(2) + ' ' + str(int_zero5).ljust(2) + ' ' 

        for ii in range( len(float_vec3) ):
            string         = string + formatfloat(float_vec3[ii],5,3) + ' ' 
         
        for ii in range( len(int_vec6) ):
            string         = string + str(int_vec6[ii]).ljust(2) + ' ' 

        string             = string + '  ' + str(byte_zero5) + ' ' 
        
        for ii in range( len(int_vec7) ):
            string         = string + str(int_vec7[ii]).ljust(2) + ' ' 
        
        string             = string + '\n' 

        return string

    else :
        # Regular animation.
        short_zero2        = footerdata[14]                # Index 14.
        int_something      = footerdata[15]                # Index 15.
        int_vec1           = footerdata[16]                # Index 16.
        int_vec2           = footerdata[17]                # Index 17.
        int_vec3           = footerdata[18]                # Index 18.
        int_vec4           = footerdata[19]                # Index 19.
        int_vec5           = footerdata[20]                # Index 20.
        int_vec6           = footerdata[21]                # Index 21.

        # Form the return string.
        string             = string + str(short_zero2).ljust(2) + ' ' + str(int_something).ljust(2) + ' ' 
        
        for ii in range( len(int_vec1) ):
            string         = string + str(int_vec1[ii]).ljust(2) + ' ' 
        
        for ii in range( len(int_vec2) ):
            string         = string + str(int_vec2[ii]).ljust(2) + ' ' 
        
        for ii in range( len(int_vec3) ):
            string         = string + str(int_vec3[ii]).ljust(2) + ' ' 
        
        for ii in range( len(int_vec4) ):
            string         = string + str(int_vec4[ii]).ljust(2) + ' ' 
        
        for ii in range( len(int_vec5) ):
            string         = string + str(int_vec5[ii]).ljust(2) + ' ' 
        
        if int_vec6 != [] :
            for ii in range( len(int_vec6) ):
                string     = string + str(int_vec6[ii]).ljust(2) + ' ' 
        
        string             = string + '\n' 

        return string

    return string


# -----------------------------------------------------------------------------------
#                            makehierarchydatastring()                              |
# -----------------------------------------------------------------------------------
def makehierarchydatastring( hierarchydata ) :

    string                 = ''
    for ii in range( len( hierarchydata ) ) :
        string             = string + str( hierarchydata[ii] ) + ' '

    string                 = string + '\n' 

    return string


# -----------------------------------------------------------------------------------
#                            animmerge()                                            |
# -----------------------------------------------------------------------------------
#                                                                                   |
#    This application prompts the user for an .ms3d file to use as the base for     |
# testing an animation and then prompts for a .cas file.  The animation is obtained |
# from the cas file, converted to Milkshape format, and put int the animation       |
# frames of the joints section.                                                     |
#                                                                                   |
# -----------------------------------------------------------------------------------
def animmerge() :
    # Construct a flags object to pass.
    flags                  = dataflags()
    flags.header           = 0
    flags.filesize         = 0
    flags.hierarchy        = 0
    flags.timeticks        = 0
    flags.bones            = 0
    flags.alldata          = 0
    flags.footer           = 0
    flags.isdir            = 0
    
    # --------------------------------------------
    # Get a ms3d filename via a GUI.             |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfnms3d             = d.go(os.curdir, "*.ms3d")
    if fullfnms3d == None :
        exit()

    # Split full filename into its three parts.
    [ pms3d, fms3d, ems3d] = splitpath( fullfnms3d )


    # --------------------------------------------
    # Get a cas filename via a GUI.              |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfncas              = d.go(os.curdir, "*.cas")
    if fullfncas == None :
        exit()

    # Split full filename into its three parts.
    [ pcas, fcas, ecas ]   = splitpath( fullfncas )


    # --------------------------------------------
    # Get desired FPS via a GUI.                 |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    FPS                    = askinteger( "Frames Per Second Dialog", "Input desired Frames Per Second (20 max, 1 min)", minvalue = 1, maxvalue = 20 )
    
    if FPS == 0 :
        exit()

    # Calculate the time step value for the joints animation entry.
    time_step              = 1.0 / float( FPS )       


    # --------------------------------------------
    # Create output file names.                  |
    # --------------------------------------------

    # Create a text file name to hold the debug results of reading the ms3dfile.
    fnms3d_debug           = pms3d + '\\' + fms3d + '_debug.txt'

    # Create a combined file name to hold the animation.  Make it a concatenation of 
    # .ms3d file name and .cas file name for uniqueness/
    fnms3d_anim            = pms3d + '\\' + fms3d + '_animby_' + fcas + '.ms3d'

    # Read the base .ms3d file that we are going to merge an animation .cas file with.
    ms3dfiledatalist       = readms3dfile( fullfnms3d, fnms3d_debug )

    # Read the .cas file that we are going to get animation data from.
    # Note that we pass in [] for the text file so we don't get a .txt representation of the .cas file.
#    fnDEBUG                = 'defend_in.txt'
#    casfiledatalist        = readcasfile( fullfncas, fnDEBUG, flags )
    casfiledatalist        = readcasfile( fullfncas, [], flags )

    # Print file names and data sizes.
    print 'fnms3d:       the .ms3d file we read from           ' + fullfnms3d
    print 'fncas:        the .cas  file we read from           ' + fullfncas
    print 'fnms3d_debug: the .txt  file we write debug info to ' + fnms3d_debug
    print 'fnms3d_anim:  the .ms3d file we write animations to ' + fnms3d_anim
    print 'length of data list = ' + str( len( ms3dfiledatalist ) )
    print 'length of data list from cas file = ' + str( len( casfiledatalist ) )


    # --------------------------------------------
    # Extract relevent data from both files.     |
    # --------------------------------------------

    # Here's all we need from the .cas file.
    headerdata             = casfiledatalist[0]
    hierarchydata          = casfiledatalist[3]
    timeticksdata          = casfiledatalist[4]
    bonedata               = casfiledatalist[5]
    animfloats_temp        = casfiledatalist[7]            # _temp because they might be out of order or missing.
    posefloats_temp        = casfiledatalist[8]            # _temp because they might be out of order or missing. 
    eulers_temp            = casfiledatalist[9]            # _temp because they might be out of order or missing. 
    footerdata             = casfiledatalist[10]
    
    nbones                 = len( hierarchydata ) - 1      # For merging we don't use Scene_Root.
    nframes                = len( timeticksdata )

    # Extract this bone's cas file bone data.
    quatframesperbone_temp = bonedata[1]                   # _temp because they might be out of order or missing. 
    animframesperbone_temp = bonedata[2]                   # _temp because they might be out of order or missing. 

    # Here's what we need from the .ms3d file.
    keyframerdata          = ms3dfiledatalist[5]
    jointdata              = ms3dfiledatalist[6]

    # -------------------------------------------------------------------------------
    #    Here's our bow to zxiang1983.  Led the effort to bring in new animations   |                 
    # and uncovered the non-standard skeleton order in new_2h_axe_halberd directory.|
    # -------------------------------------------------------------------------------

    # Do we need to re-order the .cas data?  (Non-standard skeleton order.)
    njoints                = len( jointdata )
    ms3dbonenames          = []
    for ii in range( njoints ) :
        joint              = jointdata[ii]
        ms3dbonenames.append( joint[1] )

    casbonenames           = bonedata[0]
    reorderindices         = []
    reorderindices.append( 0 )                             # Does Scene_Root.
    for ii in range( njoints ) :
        ms3dname           = ms3dbonenames[ii] 
        print "ms3dname = " + ms3dname
        for jj in range( len( casbonenames ) ) :
            casname        = casbonenames[jj] 
            print "casname = " + casname
            if casname.lower() == ms3dname.lower() :
                reorderindices.append( jj )

    print "Number of reorderindices = " + str( len( reorderindices ) )
    for ii in range( len( reorderindices ) ) :
        print "ii = " + str( ii ) + " reorderindices[ii] = " + str( reorderindices[ii] ) 

    # reorderindices has the right number to use to point to the correct data
    # so we use it to form our offsets to get at the quat data, anim data, and
    # the skeleton pose data.
    quatframesperbone      = array.array( 'I' )
    animframesperbone      = array.array( 'I' )
    for ii in range( nbones+1 ) :
        quatframesperbone.append( quatframesperbone_temp[reorderindices[ii]] )
        animframesperbone.append( animframesperbone_temp[reorderindices[ii]] )


    # Now reorder the data.  
    animfloats             = array.array( 'f' )
    posefloats             = array.array( 'f' )
    eulers                 = array.array( 'f' )
    for ibone in range( nbones) :
        if quatframesperbone[ibone+1] != 0 :
            idxeuleroffset  = 0
            for ii in range( reorderindices[ibone+1] ) :
                idxeuleroffset = idxeuleroffset + quatframesperbone_temp[ii] * 3
            for iframe in range( quatframesperbone[ibone+1] ) :
                idx        = idxeuleroffset + iframe * 3
                eulers.append( eulers_temp[idx+0] )
                eulers.append( eulers_temp[idx+1] )
                eulers.append( eulers_temp[idx+2] )

#        print 'ibone = ' + str( ibone ) + ', idxeuleroffset = ' + str( idxeuleroffset )

        if animframesperbone[ibone+1] != 0 :
            idxanimoffset  = 0
            for ii in range( reorderindices[ibone+1] ) :
                idxanimoffset = idxanimoffset + animframesperbone_temp[ii] * 3
            for iframe in range( animframesperbone[ibone+1] ) :
                idx        = idxanimoffset + iframe * 3
                animfloats.append( animfloats_temp[idx+0] )
                animfloats.append( animfloats_temp[idx+1] )
                animfloats.append( animfloats_temp[idx+2] )

#        print 'ibone = ' + str( ibone ) + ', idxanimoffset = ' + str( idxanimoffset )

    for ibone in range( nbones+1 ) :
        posefloats.append( posefloats_temp[3*reorderindices[ibone]+0] )
        posefloats.append( posefloats_temp[3*reorderindices[ibone]+1] )
        posefloats.append( posefloats_temp[3*reorderindices[ibone]+2] )

    # DEBUG STUFF.
#    print 'Size of animfloats ' + str( len( animfloats ) )
#    print 'Size of posefloats ' + str( len( posefloats ) )
#    print 'Size of eulers     ' + str( len( eulers ) )    
#    print 'nframes            ' + str( nframes )          
#    print 'nbones             ' + str( nbones )          
#    for ii in range( len( reorderindices ) ) :
#        print 'reorderindices = ' + str( reorderindices[ii] )

    # Keyframer mods.
    keyframerdata[0]       = FPS                           # User entered value.     
    keyframerdata[1]       = 0.0                           # fAnimationTime.
    keyframerdata[2]       = nframes                       # Number of frames.
    ms3dfiledatalist[5]    = keyframerdata                 # Stuff back in the new keyframer data.

    # Do the animation frames.
    quatbonecount          = 0
    animbonecount          = 0
    for ibone in range( nbones ) :
        # Extract this bone's .ms3d file joint data.
        joint              = jointdata[ibone]
        localrotation      = joint[3]
        localposition      = joint[4]

        # Make sure the local rotation is zeroed out.
        localrotation[0]   = 0.0
        localrotation[1]   = 0.0
        localrotation[2]   = 0.0
        joint[3]           = localrotation

        # Put pose data into local position.
        localposition[0]   = -posefloats[3+3*ibone+0]      # NOTE THE MINUS SIGN!
        localposition[1]   = posefloats[3+3*ibone+1]
        localposition[2]   = posefloats[3+3*ibone+2]
        joint[4]           = localposition

        # Now for the animation frames.
        rotationframes     = array.array( 'f' )
        for iframe in range( nframes ) :
            rotationframes.append( time_step * float( iframe ) )
            idx            = quatbonecount * nframes * 3 + iframe * 3
            if quatframesperbone[ibone+1] == 0 :           # NOTE: ibone+1 to jump over Scene_Root!
                rotationframes.append( 0.0 )
                rotationframes.append( 0.0 )
                rotationframes.append( 0.0 )
            else :
#                print 'idx = ' + str( idx )
                rotationframes.append( eulers[idx+0] )
                rotationframes.append( eulers[idx+1] )
                rotationframes.append( eulers[idx+2] )

        if quatframesperbone[ibone+1] != 0 :               # NOTE: ibone+1 to jump over Scene_Root! 
            quatbonecount  = quatbonecount + 1

        joint[5]           = rotationframes                # Replace rotationframes in joint[5] field.


        positionframes     = array.array( 'f' )
        for iframe in range( nframes ) :
            positionframes.append( time_step * float( iframe ) )
            idx            = animbonecount * nframes * 3 + iframe * 3
            if animframesperbone[ibone+1] == 0 :           # NOTE: ibone+1 to jump over Scene_Root! 
                positionframes.append( 0.0 )
                positionframes.append( 0.0 )
                positionframes.append( 0.0 )
            else :
                positionframes.append(-animfloats[idx+0] ) # Note the minus sign!
                positionframes.append( animfloats[idx+1] )
                positionframes.append( animfloats[idx+2] )

        if animframesperbone[ibone+1] != 0 :               # NOTE: ibone+1 to jump over Scene_Root! 
            animbonecount  = animbonecount + 1

        joint[6]           = positionframes                # Replace positionframes in joint[6] field.
        jointdata[ibone]   = joint                         # Save new joint in jointdata.

    ms3dfiledatalist[6]    = jointdata                     # Stuff the new animations back in ms3dfiledatalist[6].

    # Now we have to make ASCII strings of the header and footer to store in 
    # the joint comments section of the .ms3d file.
    headerdatastring       = makeheaderdatastring( headerdata )
    hierarchydatastring    = makehierarchydatastring( hierarchydata )
    if footerdata != [] :
       footerdatastring    = makefooterdatastring( footerdata )
    else :
       footerdatastring    = ''
    print 'headerdatastring    = ' + headerdatastring
    print 'hierarchydatastring = ' + hierarchydatastring
    print 'footerdatastring    = ' + footerdatastring
    jointcommentdata       = []
    jointindices           = []
    jointcomments          = []
    jointindices.append( 0 )
    jointcomments.append( headerdatastring )
    jointindices.append( 1 )
    jointcomments.append( hierarchydatastring )
    if footerdatastring != '' :
        jointindices.append( 2 )
        jointcomments.append( footerdatastring )

    jointcommentdata.append( jointindices )
    jointcommentdata.append( jointcomments )

    ms3dfiledatalist[10]   = jointcommentdata
    print 'length of ms3dfiledatalist = ' + str( len( ms3dfiledatalist ) )

    # Write out the new .ms3d file with the added animation data.
    writems3dfile( fnms3d_anim, ms3dfiledatalist )

    return

# -----------------------------------------------------------------------------------
#                            getcasheaderdata()                                     |
# -----------------------------------------------------------------------------------
def getcasheaderdata( jointcommentdata, animtime ) :

    # Return value.
    headerdata             = []

    # Unpack jointindices and jointcomments and the headerdatastring.
    jointindices           = jointcommentdata[0]
    jointcomments          = jointcommentdata[1]
    headerdatastring       = jointcomments[0]

    # Tokenize the line.
    tokens                 = headerdatastring.split()

    # Allocate byte arrays for signature data.
    signaturebytetriple1   = array.array( 'b' )
    signaturebytetriple2   = array.array( 'b' )

    # 42 bytes of header here.
    float_version          = float(   tokens[0] )          #  4 bytes.
    int_thirtyeight        = int(     tokens[1] )          #  8 bytes.
    int_nine               = int(     tokens[2] )          # 12 bytes.
    int_zero1              = int(     tokens[3] )          # 16 bytes. We throw away the stored animtime. If user added frames it's wrong now.
    float_animtime         = animtime                      # 20 bytes. 
    int_one1               = int(     tokens[5] )          # 24 bytes. 
    int_zero2              = int(     tokens[6] )          # 28 bytes. 
    signaturebytetriple1.append( int( tokens[7] ) )        # 29 bytes. 
    signaturebytetriple1.append( int( tokens[8] ) )        # 30 bytes. 
    signaturebytetriple1.append( int( tokens[9] ) )        # 31 bytes. 
    int_one2               = int(     tokens[10] )         # 35 bytes. 
    int_zero3              = int(     tokens[11] )         # 39 bytes. 
    signaturebytetriple2.append( int( tokens[12] ) )       # 40 bytes. 
    signaturebytetriple2.append( int( tokens[13] ) )       # 41 bytes. 
    signaturebytetriple2.append( int( tokens[14] ) )       # 42 bytes. 

    # Form into list.
    headerdata.append( float_version )                     # Index 0.
    headerdata.append( int_thirtyeight )                   # Index 1. 
    headerdata.append( int_nine )                          # Index 2. 
    headerdata.append( int_zero1 )                         # Index 3. 
    headerdata.append( float_animtime )                    # Index 4. 
    headerdata.append( int_one1 )                          # Index 5. 
    headerdata.append( int_zero2 )                         # Index 6. 
    headerdata.append( signaturebytetriple1 )              # Index 7. 
    headerdata.append( int_one2 )                          # Index 8. 
    headerdata.append( int_zero3 )                         # Index 9. 
    headerdata.append( signaturebytetriple2 )              # Index 10. 

    return headerdata

# -----------------------------------------------------------------------------------
#                            getcashierarchytree()                                  |
# -----------------------------------------------------------------------------------
def getcashierarchytree( jointcommentdata ) :

    # Return value.
    hierarchydata          = array.array( 'I' )            # Unsigned int array.

    # Unpack jointindices and jointcomments and the hierarchydatastring.
    jointindices           = jointcommentdata[0]
    jointcomments          = jointcommentdata[1]
    hierarchydatastring    = jointcomments[1]

    # Tokenize the line.
    tokens                 = hierarchydatastring.split()

    # Append the tokens.
    ntokens                = len( tokens )
    for ii in range( ntokens ) :
        hierarchydata.append( int( tokens[ii] ) )

    return hierarchydata

# -----------------------------------------------------------------------------------
#                            getcastimeticks()                                      |
# -----------------------------------------------------------------------------------
def getcastimeticks( nframes ) :

    timeticksdata          = array.array( 'f' )            # Float array.
    for ii in range( nframes ) :
        timeticksdata.append( float( ii * 0.05 ) )

    return timeticksdata

# -----------------------------------------------------------------------------------
#                            getcasbonesection()                                    |
# -----------------------------------------------------------------------------------
def getcasbonesection( jointdata ) :

    # This is the return value.
    bonedata               = []

    # Allocate arrays.
    bonenames              = []
    quatframesperbone      = array.array( 'I' )            # Unsigned int array.
    animframesperbone      = array.array( 'I' )            # Unsigned int array.
    quatoffsetperbone      = array.array( 'I' )            # Unsigned int array.
    animoffsetperbone      = array.array( 'I' )            # Unsigned int array.
    zerosperbone           = array.array( 'I' )            # Unsigned int array.
    onesperbone            = array.array( 'I' )            # Unsigned int array.

    # This is number of bones excluding Scene_Root.  But it will have weapons/shields in it.
    nbonestemp             = len( jointdata )
    nbones                 = 0
    for ii in range( nbonestemp ) :
        joint              = jointdata[ii]
        bonename           = joint[1]
        nweapon            = bonename.find( 'weapon' )
        nshield            = bonename.find( 'shield' )
        if ( nweapon > -1 ) | ( nshield > -1 ) :
           break
        nbones             = nbones + 1
    
    # Process the jointdata to fill in quatframesperbone and animframesperbone.
    quatframesperbone.append( 0 )                          # These four are for Scene_Root.
    animframesperbone.append( 0 )
    zerosperbone.append( 0 )
    onesperbone.append( 1 )
    nquatframes            = 0
    for ii in range( nbones ) :
        joint              = jointdata[ii]
        rotationframes     = joint[5]
        positionframes     = joint[6]
        nquats             = len( rotationframes ) / 4
        nanims             = len( positionframes ) / 4
        nquatframes        = nquatframes + nquats
        quatframesperbone.append( nquats )                 
        animframesperbone.append( nanims )                    
        zerosperbone.append( 0 )
        onesperbone.append( 1 )

    animsbaseoffset        = nquatframes * 16

    # Do the Scene_Root bone separately.
    bonenames.append( 'Scene_Root' )
    quatoffsetperbone.append( 0 )
    animoffsetperbone.append( animsbaseoffset )

    nquatframes            = 0
    nanimframes            = 0
    nquats                 = 0
    nanims                 = 0
    nquatbytes             = 0
    nanimbytes             = animsbaseoffset
    for ii in range( nbones ) :
        joint              = jointdata[ii]
        bonename           = joint[1]
        rotationframes     = joint[5]
        positionframes     = joint[6]
        nquatbytes         = nquatbytes + 16 * nquats
        nanimbytes         = nanimbytes + 12 * nanims
        bonenames.append( bonename )
        quatoffsetperbone.append( nquatbytes )
        animoffsetperbone.append( nanimbytes )
        nquats             = len( rotationframes ) / 4
        nanims             = len( positionframes ) / 4
        nquatframes        = nquatframes + nquats
        nanimframes        = nanimframes + nanims

    # Form the data structure.
    bonedata.append( bonenames )
    bonedata.append( quatframesperbone )
    bonedata.append( animframesperbone )
    bonedata.append( quatoffsetperbone )
    bonedata.append( animoffsetperbone )
    bonedata.append( zerosperbone )
    bonedata.append( onesperbone )

    return bonedata

# -----------------------------------------------------------------------------------
#                            getcasdata()                                           |
# -----------------------------------------------------------------------------------
def getcasdata( jointdata ) :

    # Return value.
    casdata                = []

    # This is number of bones excluding Scene_Root.  But it will have weapons/shields in it.
    nbonestemp             = len( jointdata )
    nbones                 = 0
    for ii in range( nbonestemp ) :
        joint              = jointdata[ii]
        bonename           = joint[1]
        nweapon            = bonename.find( 'weapon' )
        nshield            = bonename.find( 'shield' )
        if ( nweapon > -1 ) | ( nshield > -1 ) :
           break
        nbones             = nbones + 1

    # Allocate arrays.
    animfloats             = array.array( 'f' )
    posefloats             = array.array( 'f' )
    eulers                 = array.array( 'f' )

    # Do the Scene_Root bit of posefloats.
    posefloats.append( 0.0 )
    posefloats.append( 0.0 )
    posefloats.append( 0.0 )

    for ii in range( nbones ) :
        joint              = jointdata[ii]
        localposition      = joint[4]
        posefloats.append(-localposition[0] )
        posefloats.append( localposition[1] )
        posefloats.append( localposition[2] )
        rotationframes     = joint[5]
        nframes            = len( rotationframes ) / 4
        for iframe in range( nframes ) :
            eulers.append( rotationframes[4*iframe+1] )    # NOTE: The offset 0 one is the frame number.
            eulers.append( rotationframes[4*iframe+2] )
            eulers.append( rotationframes[4*iframe+3] )

        positionframes     = joint[6]
        nframes            = len( positionframes ) / 4
        for iframe in range( nframes ) :
            animfloats.append(-positionframes[4*iframe+1] )# NOTE: The minus sign on the x-component.   
            animfloats.append( positionframes[4*iframe+2] )
            animfloats.append( positionframes[4*iframe+3] )

    quatfloats             = computequats( eulers )

    casdata.append( quatfloats )
    casdata.append( animfloats )
    casdata.append( posefloats )
    casdata.append( eulers )     

    return casdata

# -----------------------------------------------------------------------------------
#    Processes the .cas file footer stored in jointcomments.  If jointcomments only |
# has 2 entries then there is no footer and an empty list is returned.              |
# -----------------------------------------------------------------------------------
def getcasfooter( jointcommentdata ) :

    # This is the return value.
    footerdata             = []

    # Unpack jointindices and jointcomments.
    jointindices           = jointcommentdata[0]
    jointcomments          = jointcommentdata[1]
    ncomments              = len( jointcomments )
    if ncomments == 2 :
        return footerdata

    # Get footerdatastring.
    footerdatastring       = jointcomments[2]

    # Tokenize the line.
    tokens                 = footerdatastring.split()

    # Read the first int to decide which type of footer this is: camel, fs_horse, or regular.
    int_104                = int( tokens[0] )     

    if int_104 == 18 :
        # It's a camel type.
        int_one1           = int(   tokens[1] )
        int_zero1          = int(   tokens[2] )
        int_zero2          = int(   tokens[3] ) 
        short_zero1        = int(   tokens[4] )

        int_vec1           = array.array( 'I' )
        int_vec1.append(     int(   tokens[5] ) )
        int_vec1.append(     int(   tokens[6] ) )
        int_vec1.append(     int(   tokens[7] ) )
        int_vec1.append(     int(   tokens[8] ) )

        int_vec2           = array.array( 'I' )
        int_vec2.append(     int(   tokens[9] )  )
        int_vec2.append(     int(   tokens[10] ) )
        int_vec2.append(     int(   tokens[11] ) )
        int_vec2.append(     int(   tokens[12] ) )

        int_vec3           = array.array( 'I' )
        int_vec3.append(     int(   tokens[13] ) )
        int_vec3.append(     int(   tokens[14] ) )
        int_vec3.append(     int(   tokens[15] ) )
        int_vec3.append(     int(   tokens[16] ) )

        int_vec4           = array.array( 'I' )
        int_vec4.append(     int(   tokens[17] ) )
        int_vec4.append(     int(   tokens[18] ) )
        int_vec4.append(     int(   tokens[19] ) )
        int_vec4.append(     int(   tokens[20] ) )

        int_vec5           = array.array( 'I' )
        int_vec5.append(     int(   tokens[21] ) )
        int_vec5.append(     int(   tokens[22] ) )
        int_vec5.append(     int(   tokens[23] ) )
                                                 
        footerdata.append( int_104 )
        footerdata.append( int_one1 )
        footerdata.append( int_zero1 )
        footerdata.append( int_zero2 )
        footerdata.append( short_zero1 )
        footerdata.append( int_vec1 )
        footerdata.append( int_vec2 )
        footerdata.append( int_vec3 )
        footerdata.append( int_vec4 )
        footerdata.append( int_vec5 )

        return footerdata

    # Continue regular processing.
    int_one1               = int( tokens[1] )
    int_one2               = int( tokens[2] )
    AttribString           = tokens[3]                  

    # Now read an int and a byte, then two ints and a byte.
    int_one3               = int( tokens[4] )
    byte_zero1             = int( tokens[5] )
    int_one4               = int( tokens[6] )
    int_zero1              = int( tokens[7] )
    byte_zero2             = int( tokens[8] )

    # Now read 7 floats.
    float_vec1             = array.array( 'f' )
    float_vec1.append( float( tokens[9] ) )
    float_vec1.append( float( tokens[10] ) )
    float_vec1.append( float( tokens[11] ) )
    float_vec1.append( float( tokens[12] ) )
    float_vec1.append( float( tokens[13] ) )
    float_vec1.append( float( tokens[14] ) )
    float_vec1.append( float( tokens[15] ) )

    # Now read an int and a short.
    int_zero2              = int( tokens[16] )
    short_zero1            = int( tokens[17] )

    # Now read a -1 int and another int zero.
    int_minus1             = int( tokens[18] )  
    int_zero3              = int( tokens[19] )

    # Save off the common data.
    footerdata.append( int_104 )                           # Index 0.
    footerdata.append( int_one1 )                          # Index 1.
    footerdata.append( int_one2 )                          # Index 2.
    footerdata.append( AttribString )                      # Index 3. (Remember the null terminator byte.)
    footerdata.append( int_one3 )                          # Index 4.
    footerdata.append( byte_zero1 )                        # Index 5.
    footerdata.append( int_one4 )                          # Index 6.
    footerdata.append( int_zero1 )                         # Index 7.
    footerdata.append( byte_zero2 )                        # Index 8.
                                             
    footerdata.append( float_vec1 )                        # Index 9.
                                             
    footerdata.append( int_zero2 )                         # Index 10.
    footerdata.append( short_zero1 )                       # Index 11.
    footerdata.append( int_minus1 )                        # Index 12.
    footerdata.append( int_zero3 )                         # Index 13.

    if int_104 == 170 :
        # Doing the horsies. 
        string2            = tokens[20]                 
        int_one5           = int( tokens[21] )  
        byte_zero3         = int( tokens[22] )  
        int_one6           = int( tokens[23] )
        int_zero4          = int( tokens[24] )
        byte_zero4         = int( tokens[25] )

        # 7 floats.
        float_vec2         = array.array( 'f' )
        float_vec2.append( float( tokens[26] ) )
        float_vec2.append( float( tokens[27] ) )
        float_vec2.append( float( tokens[28] ) )
        float_vec2.append( float( tokens[29] ) )
        float_vec2.append( float( tokens[30] ) )
        float_vec2.append( float( tokens[31] ) )
        float_vec2.append( float( tokens[32] ) )

        int_vec1           = array.array( 'I' )
        int_vec1.append( int( tokens[33] ) )
        int_vec1.append( int( tokens[34] ) )
        int_vec1.append( int( tokens[35] ) )
        int_vec1.append( int( tokens[36] ) )
        int_vec1.append( int( tokens[37] ) )

        int_vec2           = array.array( 'I' )
        int_vec2.append( int( tokens[38] ) )
        int_vec2.append( int( tokens[39] ) )
        int_vec2.append( int( tokens[40] ) )
        int_vec2.append( int( tokens[41] ) )

        int_vec3           = array.array( 'I' )
        int_vec3.append( int( tokens[42] ) )
        int_vec3.append( int( tokens[43] ) )
        int_vec3.append( int( tokens[44] ) )
        int_vec3.append( int( tokens[45] ) )

        int_vec4           = array.array( 'I' )
        int_vec4.append( int( tokens[46] ) )
        int_vec4.append( int( tokens[47] ) )
        int_vec4.append( int( tokens[48] ) )
        int_vec4.append( int( tokens[49] ) )

        int_vec5           = array.array( 'I' )
        int_vec5.append( int( tokens[50] ) )
        int_vec5.append( int( tokens[51] ) )
        int_vec5.append( int( tokens[52] ) )
        int_vec5.append( int( tokens[53] ) )

        int_count          = int( tokens[54] )             # Byte count of what is coming.
        int_five           = int( tokens[55] )
        int_one7           = int( tokens[56] )
        int_zero5          = int( tokens[57] )

        # 7 more floats.
        float_vec3         = array.array( 'f' )
        float_vec3.append( float( tokens[58] ) )
        float_vec3.append( float( tokens[59] ) )
        float_vec3.append( float( tokens[60] ) )
        float_vec3.append( float( tokens[61] ) )
        float_vec3.append( float( tokens[62] ) )
        float_vec3.append( float( tokens[63] ) )
        float_vec3.append( float( tokens[64] ) )
        float_vec3.fromfile( fidcas, 7 )

        # 7 ints.
        int_vec6           = array.array( 'I' )
        int_vec6.append( int( tokens[65] ) )
        int_vec6.append( int( tokens[66] ) )
        int_vec6.append( int( tokens[67] ) )
        int_vec6.append( int( tokens[68] ) )
        int_vec6.append( int( tokens[69] ) )
        int_vec6.append( int( tokens[70] ) )
        int_vec6.append( int( tokens[71] ) )

        byte_zero5         = int( tokens[72] )  

        # Should be last 12 12 0 ints.
        int_vec7           = array.array( 'I' )
        int_vec7.append( int( tokens[73] ) )
        int_vec7.append( int( tokens[74] ) )
        int_vec7.append( int( tokens[75] ) )

        # Save off the data.
        footerdata.append( string2 )                       # Index 14.
        footerdata.append( int_one5 )                      # Index 15.
        footerdata.append( byte_zero3 )                    # Index 16.
        footerdata.append( int_one6 )                      # Index 17.
        footerdata.append( int_zero4 )                     # Index 18.
        footerdata.append( byte_zero4 )                    # Index 19.
        footerdata.append( float_vec2 )                    # Index 20.
        footerdata.append( int_vec1 )                      # Index 21.
        footerdata.append( int_vec2 )                      # Index 22.
        footerdata.append( int_vec3 )                      # Index 23.
        footerdata.append( int_vec4 )                      # Index 24.
        footerdata.append( int_vec5 )                      # Index 25.
        footerdata.append( int_count )                     # Index 26.
        footerdata.append( int_five )                      # Index 27.
        footerdata.append( int_one7 )                      # Index 28.
        footerdata.append( int_zero5 )                     # Index 29.
        footerdata.append( float_vec3 )                    # Index 30.
        footerdata.append( int_vec6 )                      # Index 31.
        footerdata.append( byte_zero5 )                    # Index 32.
        footerdata.append( int_vec7 )                      # Index 33.

        return footerdata

    else :
        short_zero2        = int( tokens[20] )
        int_something      = int( tokens[21] )

        # Now read a long bit of 19 ints.
        int_vec1           = array.array( 'I' )
        int_vec1.append( int( tokens[22] ) )
        int_vec1.append( int( tokens[23] ) )
        int_vec1.append( int( tokens[24] ) )
        int_vec1.append( int( tokens[25] ) )

        int_vec2           = array.array( 'I' )
        int_vec2.append( int( tokens[26] ) )
        int_vec2.append( int( tokens[27] ) )
        int_vec2.append( int( tokens[28] ) )
        int_vec2.append( int( tokens[29] ) )

        int_vec3           = array.array( 'I' )
        int_vec3.append( int( tokens[30] ) )
        int_vec3.append( int( tokens[31] ) )
        int_vec3.append( int( tokens[32] ) )
        int_vec3.append( int( tokens[33] ) )

        int_vec4           = array.array( 'I' )
        int_vec4.append( int( tokens[34] ) )
        int_vec4.append( int( tokens[35] ) )
        int_vec4.append( int( tokens[36] ) )
        int_vec4.append( int( tokens[37] ) )

        int_vec5           = array.array( 'I' )
        int_vec5.append( int( tokens[38] ) )
        int_vec5.append( int( tokens[39] ) )
        int_vec5.append( int( tokens[40] ) )

        # MAYBE read 3 ints if not a bad file.
        if len( tokens ) == 41 :
            int_vec6       = []
        else :
            int_vec6       = array.array( 'I' )
            int_vec6.append( int( tokens[41] ) )
            int_vec6.append( int( tokens[42] ) )
            int_vec6.append( int( tokens[43] ) )

        # Save off the data pertaining to non fs_horse.
        footerdata.append( short_zero2 )                   # Index 14.
        footerdata.append( int_something )                 # Index 15.
        footerdata.append( int_vec1 )                      # Index 16.
        footerdata.append( int_vec2 )                      # Index 17.
        footerdata.append( int_vec3 )                      # Index 18.
        footerdata.append( int_vec4 )                      # Index 19.
        footerdata.append( int_vec5 )                      # Index 20.
        footerdata.append( int_vec6 )                      # Index 21.

        return footerdata

    return footerdata


# -----------------------------------------------------------------------------------
#                            animextract()                                          |
# -----------------------------------------------------------------------------------
#                                                                                   |
#    This application prompts the user for an .ms3d file from which it is desired   |
# to extract an animation.  The user is responsible for knowing that the file       |
# actually HAS an animation to extract.  It then extracts the animation frames and  |
# writes a full format (with header and footer) .cas file.                          |
#                                                                                   |
# -----------------------------------------------------------------------------------
def animextract() :

    # --------------------------------------------
    # Get an animated ms3d filename via a GUI.   |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfnms3d             = d.go(os.curdir, "*.ms3d")
    if fullfnms3d == None :
        exit()

    # Split full file name into three parts.
    [ pms3d, fms3d, ems3d ]= splitpath( fullfnms3d )

    # Get the .cas file name to use from the second half of the animated .ms3d file name.
    nanimby                = fms3d.find( '_animby_' )
    fncas                  = pms3d + '\\' + fms3d[nanimby+8:] + '_modified.cas'

    # Make a debug text name.
    fndebugtxt             = pms3d + '\\' + fms3d + '_debug.txt'

    # --------------------------------------------
    # Read the .ms3d file to get the animations. |
    # --------------------------------------------
    ms3dfiledatalist       = readms3dfile( fullfnms3d, fndebugtxt )

    # --------------------------------------------
    # Extract info needed to write a cas file.   |
    # --------------------------------------------
    jointdata              = ms3dfiledatalist[6]
    jointcommentdata       = ms3dfiledatalist[10]
    joint0                 = jointdata[0]
    rotationframes         = joint0[6]

    nframes                = len( rotationframes ) / 4     # NOTE: This is hard coding...won't work with flailing .cas's.
    animtime               = nframes * 0.05
    headerdata             = getcasheaderdata( jointcommentdata, animtime )

    filesizesans           = 0                             # NOTE: This is recomputed by writecasfile().
    
    hierarchydata          = getcashierarchytree( jointcommentdata ) 

    timeticksdata          = getcastimeticks( nframes ) 

    bonedata               = getcasbonesection( jointdata ) 

    casdata                = getcasdata( jointdata ) 
    quatfloats             = casdata[0]
    animfloats             = casdata[1]
    posefloats             = casdata[2]
    eulers                 = casdata[3]

    footerdata             = getcasfooter( jointcommentdata ) 

    # Make the casfiledatalist structure.
    casfiledatalist        = []
    casfiledatalist.append( headerdata )
    casfiledatalist.append( filesizesans )
    casfiledatalist.append( 0 )           
    casfiledatalist.append( hierarchydata )
    casfiledatalist.append( timeticksdata )
    casfiledatalist.append( bonedata )        
    casfiledatalist.append( quatfloats )      
    casfiledatalist.append( animfloats )      
    casfiledatalist.append( posefloats )      
    casfiledatalist.append( eulers )          
    casfiledatalist.append( footerdata )      

    # Use the library function.
    writecasfile( fncas, casfiledatalist )

    print 'Animation extracted to: ' + fncas

    return


# -----------------------------------------------------------------------------------
#                            Copied from the Python 2.5 lib.  Have to change one    |
#                            thing to get it to work right.                         |
# -----------------------------------------------------------------------------------


#
# An Introduction to Tkinter
# tkSimpleDialog.py
#
# Copyright (c) 1997 by Fredrik Lundh
#
# fredrik@pythonware.com
# http://www.pythonware.com
#

# --------------------------------------------------------------------
# dialog base class

'''Dialog boxes

This module handles dialog boxes. It contains the following
public symbols:

Dialog -- a base class for dialogs

askinteger -- get an integer from the user

askfloat -- get a float from the user

askstring -- get a string from the user
'''


class Dialog(Toplevel):

    '''Class to open dialogs.

    This class is intended as a base class for custom dialogs
    '''

    def __init__(self, parent, title = None):

        '''Initialize a dialog.

        Arguments:

            parent -- a parent window (the application window)

            title -- the dialog title
        '''
        Toplevel.__init__(self, parent)
#        self.transient(parent)

        if title:
            self.title(title)

        self.parent = parent

        self.result = None

        body = Frame(self)
        self.initial_focus = self.body(body)
        body.pack(padx=5, pady=5)

        self.buttonbox()

        self.wait_visibility() # window needs to be visible for the grab
        self.grab_set()

        if not self.initial_focus:
            self.initial_focus = self

        self.protocol("WM_DELETE_WINDOW", self.cancel)

        if self.parent is not None:
            self.geometry("+%d+%d" % (parent.winfo_rootx()+50,
                                      parent.winfo_rooty()+50))

        self.initial_focus.focus_set()

        self.wait_window(self)

    def destroy(self):
        '''Destroy the window'''
        self.initial_focus = None
        Toplevel.destroy(self)

    #
    # construction hooks

    def body(self, master):
        '''create dialog body.

        return widget that should have initial focus.
        This method should be overridden, and is called
        by the __init__ method.
        '''
        pass

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = Frame(self)

        w = Button(box, text="OK", width=10, command=self.ok, default=ACTIVE)
        w.pack(side=LEFT, padx=5, pady=5)
        w = Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    #
    # standard button semantics

    def ok(self, event=None):

        if not self.validate():
            self.initial_focus.focus_set() # put focus back
            return

        self.withdraw()
        self.update_idletasks()

        self.apply()

        self.cancel()

    def cancel(self, event=None):

        # put focus back to the parent window
        if self.parent is not None:
            self.parent.focus_set()
        self.destroy()

    #
    # command hooks

    def validate(self):
        '''validate the data

        This method is called automatically to validate the data before the
        dialog is destroyed. By default, it always validates OK.
        '''

        return 1 # override

    def apply(self):
        '''process the data

        This method is called automatically to process the data, *after*
        the dialog is destroyed. By default, it does nothing.
        '''

        pass # override


# --------------------------------------------------------------------
# convenience dialogues

class _QueryDialog(Dialog):

    def __init__(self, title, prompt,
                 initialvalue=None,
                 minvalue = None, maxvalue = None,
                 parent = None):

        if not parent:
            import Tkinter
            parent = Tkinter._default_root

        self.prompt   = prompt
        self.minvalue = minvalue
        self.maxvalue = maxvalue

        self.initialvalue = initialvalue

        Dialog.__init__(self, parent, title)

    def destroy(self):
        self.entry = None
        Dialog.destroy(self)

    def body(self, master):

        w = Label(master, text=self.prompt, justify=LEFT)
        w.grid(row=0, padx=5, sticky=W)

        self.entry = Entry(master, name="entry")
        self.entry.grid(row=1, padx=5, sticky=W+E)

        if self.initialvalue:
            self.entry.insert(0, self.initialvalue)
            self.entry.select_range(0, END)

        return self.entry

    def validate(self):

        import tkMessageBox

        try:
            result = self.getresult()
        except ValueError:
            tkMessageBox.showwarning(
                "Illegal value",
                self.errormessage + "\nPlease try again",
                parent = self
            )
            return 0

        if self.minvalue is not None and result < self.minvalue:
            tkMessageBox.showwarning(
                "Too small",
                "The allowed minimum value is %s. "
                "Please try again." % self.minvalue,
                parent = self
            )
            return 0

        if self.maxvalue is not None and result > self.maxvalue:
            tkMessageBox.showwarning(
                "Too large",
                "The allowed maximum value is %s. "
                "Please try again." % self.maxvalue,
                parent = self
            )
            return 0

        self.result = result

        return 1


class _QueryInteger(_QueryDialog):
    errormessage = "Not an integer."
    def getresult(self):
        return int(self.entry.get())

def askinteger(title, prompt, **kw):
    '''get an integer from the user

    Arguments:

        title -- the dialog title
        prompt -- the label text
        **kw -- see SimpleDialog class

    Return value is an integer
    '''
    d = _QueryInteger(title, prompt, **kw)
    return d.result

class _QueryFloat(_QueryDialog):
    errormessage = "Not a floating point value."
    def getresult(self):
        return float(self.entry.get())

def askfloat(title, prompt, **kw):
    '''get a float from the user

    Arguments:

        title -- the dialog title
        prompt -- the label text
        **kw -- see SimpleDialog class

    Return value is a float
    '''
    d = _QueryFloat(title, prompt, **kw)
    return d.result

class _QueryString(_QueryDialog):
    def __init__(self, *args, **kw):
        if kw.has_key("show"):
            self.__show = kw["show"]
            del kw["show"]
        else:
            self.__show = None
        _QueryDialog.__init__(self, *args, **kw)

    def body(self, master):
        entry = _QueryDialog.body(self, master)
        if self.__show is not None:
            entry.configure(show=self.__show)
        return entry

    def getresult(self):
        return self.entry.get()

def askstring(title, prompt, **kw):
    '''get a string from the user

    Arguments:

        title -- the dialog title
        prompt -- the label text
        **kw -- see SimpleDialog class

    Return value is a string
    '''
    d = _QueryString(title, prompt, **kw)
    return d.result

#if __name__ == "__main__":
#
#    root = Tk()
#    root.update()
#
#    print askinteger("Spam", "Egg count", initialvalue=12*12)
#    print askfloat("Spam", "Egg weight\n(in tons)", minvalue=1, maxvalue=100)
#    print askstring("Spam", "Egg label")

# -----------------------------------------------------------------------------------
#                            getscalingfactors()                                    |
# -----------------------------------------------------------------------------------
class getscalingfactors(Dialog) :

    def body( self, master ) :
        self.title( "Input Scaling Values" )

#        SCALEX             = 0.69466         
#        SCALEY             = 0.64000
#        SCALEZ             = 0.69466

        SCALEX             = 0.64            
        SCALEY             = 0.64   
        SCALEZ             = 0.64   

        Label( master, text = "X Scaling Factor:" ).grid( row = 0, column = 0, padx = 5, pady = 10, sticky = W )
        Label( master, text = "Y Scaling Factor:" ).grid( row = 1, column = 0, padx = 5, pady = 10, sticky = W )
        Label( master, text = "Z Scaling Factor:" ).grid( row = 2, column = 0, padx = 5, pady = 10, sticky = W )
        
        self.xstring       = Entry( master, width = 10 )
        self.ystring       = Entry( master, width = 10 )
        self.zstring       = Entry( master, width = 10 )

        self.xstring.grid( row = 0, column = 1, padx = 5, pady = 10 ) 
        self.ystring.grid( row = 1, column = 1, padx = 5, pady = 10 ) 
        self.zstring.grid( row = 2, column = 1, padx = 5, pady = 10 ) 

        self.xstring.insert( 0, str( SCALEX ) )
        self.ystring.insert( 0, str( SCALEY ) )
        self.zstring.insert( 0, str( SCALEZ ) )

        return

    def apply(self) :
        xstring            = self.xstring.get()
        ystring            = self.ystring.get()
        zstring            = self.zstring.get()
        self.xval          = float( xstring )
        self.yval          = float( ystring )
        self.zval          = float( zstring )
        return 

    def getvalues(self) :
        factors            = []
        factors.append( self.xval )
        factors.append( self.yval )
        factors.append( self.zval )
        return factors
    

# -----------------------------------------------------------------------------------
#                            exportskeleton()                                       |
# -----------------------------------------------------------------------------------
#                                                                                   |
#    This application prompts the user for a .skeleton file to export.  It then     |
# prompts the user to mouse to a directory he wants to convert and choose any file  |
# there.  This lets us get the directory name which we then process.                |
#                                                                                   |
# -----------------------------------------------------------------------------------
def exportskeleton() :

    # --------------------------------------------
    # Construct a flags object to pass.          |
    # --------------------------------------------
    flags                  = dataflags()
    flags.header           = 0
    flags.filesize         = 0
    flags.hierarchy        = 0
    flags.timeticks        = 0
    flags.bones            = 0
    flags.alldata          = 0
    flags.footer           = 0
    flags.isdir            = 0

    # --------------------------------------------
    # Get a skeleton filename via a GUI.         |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfnskeleton         = d.go(os.curdir, "*.skeleton")
    if fullfnskeleton == None :
        exit()

    # --------------------------------------------
    # Get any cas filename in directory to be    |
    # processed.                                 |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfncas              = d.go(os.curdir, "*.cas")
    if fullfncas == None :
        exit()

    # Split full filename into its three parts.
    [ pdir, fdir, edir ]   = splitpath( fullfncas )

    # --------------------------------------------
    # Get the scaling factors from the user.     |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    dialog                 = getscalingfactors(root)
    factors                = dialog.getvalues()

    SCALEX                 = factors[0] 
    SCALEY                 = factors[1] 
    SCALEZ                 = factors[2] 
    print 'SCALEX, Y, Z = ' + str( SCALEX ) + '  ' + str( SCALEY ) + ' ' + str( SCALEZ )

    # --------------------------------------------
    # Read the skeleton data into newposefloats. |
    # --------------------------------------------
    fidskel                = open( fullfnskeleton, 'r' )
    newposefloats          = array.array( 'f' )
    newposefloats.append( 0.0 )                            # These are Scene_Root poses.
    newposefloats.append( 0.0 )
    newposefloats.append( 0.0 )
    while 1 == 1 :
        line               = fidskel.readline()
        tokens             = line.split()
        if len( tokens ) > 1 :
            newposefloats.append( float( tokens[0] ) )
            newposefloats.append( float( tokens[1] ) )
            newposefloats.append( float( tokens[2] ) )
        else :
            break

    fidskel.close()

    # Need to enumerate all the .cas files in the directory pointed at by the path pdir.
    namelist               = os.listdir( pdir )
    nfiles                 = len( namelist )

    # Create a subdirectory to hold the rewritten files.
    pconvdir               = pdir + '\\convertedfiles\\'
    direxists              = os.path.isdir( pconvdir )
    if direxists == False :
        os.makedirs( pconvdir )

    # --------------------------------------------
    # Iterate over the namelist and put the new  |
    # pose data over the old and write out the   |
    # cas files.                                 |
    # --------------------------------------------
    for ifile in namelist :
        BASEPOSEFLAG       = False
        FLAILINGFLAG       = False
        print 'processing file = ' + ifile
        [ pcas, fcas, ecas ] = splitpath( ifile )       

        nbasepose          = fcas.find( 'basepose' )
        if nbasepose > -1:
            BASEPOSEFLAG   = True
            print 'We have the basepose file....'

        nflailingname  = fcas.find( 'die_flailing' )
        if nflailingname > -1 :
            FLAILINGFLAG   = True
            print 'We have the die_flailing file....'

        if ecas == '.cas' :
            fnin           = pdir + '\\' + fcas + ecas
            fnout          = pconvdir    + fcas + ecas
            #print 'fnin  = ' + fnin
            #print 'fnout = ' + fnout

            # Get data of original cas file.
            casfiledatalist = readcasfile( fnin, [], flags )

            # Substitute our new skeleton.
            casfiledatalist[8] = newposefloats

            # Actually SCALE the data this time you idiot!
            animfloats         = casfiledatalist[7]
            nfloats            = len( animfloats )
            nvectors           = nfloats / 3
            for ii in range( nvectors ) :
                animfloats[3*ii+0] = SCALEX * animfloats[3*ii+0]
                animfloats[3*ii+1] = SCALEY * animfloats[3*ii+1]
                animfloats[3*ii+2] = SCALEZ * animfloats[3*ii+2]

            casfiledatalist[7] = animfloats

            # Write new cas file out to the \convertedfiles subdirectory.
            writecasfile( fnout, casfiledatalist )

    print '  '
    print 'exportskeleton has completed processing all the .cas files in this directory'

    return

   
# -----------------------------------------------------------------------------------
#                            gethumanhierarchynumber()                              |
# -----------------------------------------------------------------------------------
def gethumanhierarchynumber( bonename ) :

    if bonename.lower() == 'bone_pelvis' :
        hierarchynumber    = 0
    elif bonename.lower() == 'bone_rthigh' :
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_rlowerleg' :
        hierarchynumber    = 2
    elif bonename.lower() == 'bone_rfoot' :
        hierarchynumber    = 3
    elif bonename.lower() == 'bone_abs' :
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_torso' :
        hierarchynumber    = 5
    elif bonename.lower() == 'bone_head' :
        hierarchynumber    = 6
    elif bonename.lower() == 'bone_jaw' :
        hierarchynumber    = 7
    elif bonename.lower() == 'bone_eyebrow' :
        hierarchynumber    = 7
    elif bonename.lower() == 'bone_rclavical' :
        hierarchynumber    = 6
    elif bonename.lower() == 'bone_rupperarm' :
        hierarchynumber    = 10
    elif bonename.lower() == 'bone_relbow' :
        hierarchynumber    = 11
    elif bonename.lower() == 'bone_rhand' :
        hierarchynumber    = 12
    elif bonename.lower() == 'bone_lclavical' :
        hierarchynumber    = 6
    elif bonename.lower() == 'bone_lupperarm' :
        hierarchynumber    = 14
    elif bonename.lower() == 'bone_lelbow' :
        hierarchynumber    = 15
    elif bonename.lower() == 'bone_lhand' :
        hierarchynumber    = 16
    elif bonename.lower() == 'bone_lthigh' :
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_llowerleg' :
        hierarchynumber    = 18
    elif bonename.lower() == 'bone_lfoot' :
        hierarchynumber    = 19
    else:
        hierarchynumber    = 99

    return hierarchynumber

   
# -----------------------------------------------------------------------------------
#                            getcamelhierarchynumber()                              |
# -----------------------------------------------------------------------------------
def getcamelhierarchynumber( bonename ) :

    if bonename.lower()   == 'bone_camel_root' :     
        hierarchynumber    = 0
    elif bonename.lower() == 'bone_camel_f_lleg_top' :
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_camel_f_lleg_low' :
        hierarchynumber    = 2
    elif bonename.lower() == 'bone_camel_f_lfoot' :  
        hierarchynumber    = 3
    elif bonename.lower() == 'bone_camel_f_rleg_top' :
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_camel_f_rleg_low' :
        hierarchynumber    = 5
    elif bonename.lower() == 'bone_camel_f_rfoot' :  
        hierarchynumber    = 6
    elif bonename.lower() == 'bone_camel_back' :     
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_camel_r_lleg_top' :
        hierarchynumber    = 8
    elif bonename.lower() == 'bone_camel_r_lleg_mid' :
        hierarchynumber    = 9
    elif bonename.lower() == 'bone_camel_r_lleg_low' :
        hierarchynumber    = 10
    elif bonename.lower() == 'bone_camel_r_lfoot' :  
        hierarchynumber    = 11
    elif bonename.lower() == 'bone_camel_r_rleg_top' :
        hierarchynumber    = 8
    elif bonename.lower() == 'bone_camel_r_rleg_mid' :
        hierarchynumber    = 13
    elif bonename.lower() == 'bone_camel_r_rleg_low' :
        hierarchynumber    = 14
    elif bonename.lower() == 'bone_camel_r_rfoot' :  
        hierarchynumber    = 15
    elif bonename.lower() == 'bone_camel_neck_low' : 
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_camel_neck_mid' : 
        hierarchynumber    = 17
    elif bonename.lower() == 'bone_camel_neck_top' : 
        hierarchynumber    = 18
    elif bonename.lower() == 'bone_camel_head' :     
        hierarchynumber    = 19
    else:
        hierarchynumber    = 99

    return hierarchynumber

   
# -----------------------------------------------------------------------------------
#                            gethorsehierarchynumber()                              |
# -----------------------------------------------------------------------------------
def gethorsehierarchynumber( bonename ) :

    if bonename.lower()   == 'bone_h_saddle' :       
        hierarchynumber    = 0
    elif bonename.lower() == 'bone_spine' :          
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_spine1' :         
        hierarchynumber    = 2
    elif bonename.lower() == 'bone_neck' :           
        hierarchynumber    = 3
    elif bonename.lower() == 'bone_head' :           
        hierarchynumber    = 4
    elif bonename.lower() == 'bone_rightarm' :       
        hierarchynumber    = 3
    elif bonename.lower() == 'bone_rightforearm' :   
        hierarchynumber    = 6
    elif bonename.lower() == 'bone_righthand' :      
        hierarchynumber    = 7
    elif bonename.lower() == 'bone_rightfingerbase' :
        hierarchynumber    = 8
    elif bonename.lower() == 'bone_leftarm' :        
        hierarchynumber    = 3
    elif bonename.lower() == 'bone_leftforearm' :    
        hierarchynumber    = 10
    elif bonename.lower() == 'bone_lefthand' :       
        hierarchynumber    = 11
    elif bonename.lower() == 'bone_leftfingerbase' : 
        hierarchynumber    = 12
    elif bonename.lower() == 'bone_tail1' :          
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_tail2' :          
        hierarchynumber    = 14
    elif bonename.lower() == 'bone_rightupleg' :     
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_rightleg' :       
        hierarchynumber    = 16
    elif bonename.lower() == 'bone_rightfoot' :      
        hierarchynumber    = 17
    elif bonename.lower() == 'bone_righttoebase' :   
        hierarchynumber    = 18
    elif bonename.lower() == 'bone_leftupleg' :      
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_leftleg' :        
        hierarchynumber    = 20
    elif bonename.lower() == 'bone_leftfoot' :       
        hierarchynumber    = 21
    elif bonename.lower() == 'bone_lefttoebase' :    
        hierarchynumber    = 22
    else:
        hierarchynumber    = 99

    return hierarchynumber

   
# -----------------------------------------------------------------------------------
#                            getelephanthierarchynumber()                           |
# -----------------------------------------------------------------------------------
def getelephanthierarchynumber( bonename ) :

    if bonename.lower()   == 'bone_e_platform' :            
        hierarchynumber    = 0
    elif bonename.lower() == 'bone_e_back' :                 
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_e_right back thigh' :    
        hierarchynumber    = 2
    elif bonename.lower() == 'bone_e_right back lower leg' :
        hierarchynumber    = 3
    elif bonename.lower() == 'bone_e_right back foot' :     
        hierarchynumber    = 4
    elif bonename.lower() == 'bone_e_left back thigh' :     
        hierarchynumber    = 2
    elif bonename.lower() == 'bone_e_left back lower leg' :  
        hierarchynumber    = 6
    elif bonename.lower() == 'bone_e_left back foot' :       
        hierarchynumber    = 7
    elif bonename.lower() == 'bone_e_head' :                
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_e_trunk 1' :             
        hierarchynumber    = 9
    elif bonename.lower() == 'bone_e_trunk 2' :             
        hierarchynumber    = 10
    elif bonename.lower() == 'bone_e_trunk 3' :             
        hierarchynumber    = 11
    elif bonename.lower() == 'bone_e_trunk 4' :             
        hierarchynumber    = 12
    elif bonename.lower() == 'bone_e_right ear main' :      
        hierarchynumber    = 9
    elif bonename.lower() == 'bone_e_right ear tip' :       
        hierarchynumber    = 14
    elif bonename.lower() == 'bone_e_left ear main' :       
        hierarchynumber    = 9
    elif bonename.lower() == 'bone_e_left ear tip' :        
        hierarchynumber    = 16
    elif bonename.lower() == 'bone_e_right front thigh' :   
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_e_right front lower leg' :
        hierarchynumber    = 18
    elif bonename.lower() == 'bone_e_right front foot' :    
        hierarchynumber    = 19
    elif bonename.lower() == 'bone_e_left front thigh' :    
        hierarchynumber    = 1
    elif bonename.lower() == 'bone_e_left front lower leg' :
        hierarchynumber    = 21
    elif bonename.lower() == 'bone_e_left front foot' :     
        hierarchynumber    = 22
    else:
        hierarchynumber    = 99

    return hierarchynumber


# -----------------------------------------------------------------------------------
#                            extractskeleton()                                      |
# -----------------------------------------------------------------------------------
#                                                                                   |
#    This application prompts the user for a .ms3d file to extract a skeleton from. |
# It writes the extracted skeleton to the current directory and names it            |
# 'extracted.skeleton'.                                                             |
#                                                                                   |
# -----------------------------------------------------------------------------------
def extractskeleton() :

    # Flags.
    HUMANFLAG              = False
    CAMELFLAG              = False
    HORSEFLAG              = False
    ELEPHANTFLAG           = False

    # Get filename from a filechooser.
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfnms3d             = d.go(os.curdir, "*.ms3d")
    if fullfnms3d == None :
        exit()

    # Extract info in ms3d file.
    ms3dfiledatalist       = readms3dfile( fullfnms3d, [] )

    # Extract the joint data.
    jointdata              = ms3dfiledatalist[6]

    # What kind of skeleton is it?
    joint                  = jointdata[0]
    bonename               = joint[1]
    if bonename.lower() == 'bone_pelvis' :
        HUMANFLAG          = True
    elif bonename.lower() == 'bone_camel_root' :
        CAMELFLAG          = True
    elif bonename.lower() == 'bone_h_saddle' :
        HORSEFLAG          = True
    elif bonename.lower() == 'bone_e_platform' :
        ELEPHANTFLAG       = True
    else:
        print 'The skeleton in this file is not recognizable as a human, camel, horse, or elephant.' 

    # This is number of bones excluding Scene_Root.  But it will have weapons/shields in it.
    nbonestemp             = len( jointdata )
    nbones                 = 0
    hierarchydata          = array.array( 'I' )
    for ii in range( nbonestemp ) :
        joint              = jointdata[ii]
        bonename           = joint[1]
        nweapon            = bonename.find( 'weapon' )
        nshield            = bonename.find( 'shield' )
        if ( nweapon > -1 ) | ( nshield > -1 ) :
           continue
        if   HUMANFLAG == True :
            hierarchydata.append( gethumanhierarchynumber( bonename ) ) 
        elif CAMELFLAG == True :
            hierarchydata.append( getcamelhierarchynumber( bonename ) ) 
        elif HORSEFLAG == True :
            hierarchydata.append( gethorsehierarchynumber( bonename ) ) 
        elif ELEPHANTFLAG == True :
            hierarchydata.append( getelephanthierarchynumber( bonename ) ) 
        nbones             = nbones + 1

    fnout                  = 'extracted.skeleton'
    fidout                 = open( fnout, 'w' )

    ibone                  = 0
    for ii in range( nbonestemp ) :
        joint              = jointdata[ii]
        bonename           = joint[1]
        nweapon            = bonename.find( 'weapon' )
        nshield            = bonename.find( 'shield' )
        if ( nweapon > -1 ) | ( nshield > -1 ) :
           continue
        localposition      = joint[4]
        xval               = -localposition[0]
        yval               = +localposition[1]
        zval               = +localposition[2]
        string             = formatfloat( xval, 12, 10 ) + '  ' + formatfloat( yval, 12, 10 ) + '  ' + formatfloat( zval, 12, 10 ) + '  ' + str(hierarchydata[ibone]).rjust(3) + '  ' + bonename + '\n' 
        fidout.write( string )
        ibone              = ibone + 1

    fidout.close()

    return


# -----------------------------------------------------------------------------------
#                            convertcastotxt()                                      |
# -----------------------------------------------------------------------------------
#                                                                                   |
#    This application prompts the user for a .cas file to convert to a text         |
# representation.  The .txt file is created in the same directory as the .cas file  |
# with the same name, just different extension.                                     |
#                                                                                   |
# -----------------------------------------------------------------------------------
def convertcastotxt():

    # Get filename from a filechooser.
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfncas              = d.go(os.curdir, "*.cas")
    if fullfncas == None :
        exit()

    # Split full filename into its three parts.
    [ pdir, fdir, edir ]   = splitpath( fullfncas )

    # Make an output .txt filename.
    fullfntxt              = pdir + '\\' + fdir + '.txt'

    # Set the flags to write all the data and set flags.isdir = 0.
    flags                  = dataflags()
    flags.header           = 1
    flags.filesize         = 1
    flags.hierarchy        = 1
    flags.timeticks        = 1
    flags.bones            = 1
    flags.alldata          = 1
    flags.footer           = 1
    flags.isdir            = 0
    
    # Call readcasfile.  The side effect of this with flags set as above and a valid text filename 
    # is that the function writes out a full text representation of the cas file.
    casfiledatalist        = readcasfile( fullfncas, fullfntxt, flags )
    print 'Converted file ' + fdir + edir + ' to text as file ' + fdir + '.txt in directory ' + pdir

    return

# -----------------------------------------------------------------------------------
#                            dataflags() - Equivalent to a C struct.                |
# -----------------------------------------------------------------------------------
class dataflags() :

    def __init__(self) :
        self.header        = 0
        self.filesize      = 0
        self.hierarchy     = 0
        self.timeticks     = 0
        self.bones         = 0
        self.alldata       = 0
        self.footer        = 0
        self.isdir         = 0

        return


# -----------------------------------------------------------------------------------
#                            getdataoptions()                                       |
# -----------------------------------------------------------------------------------
class getdataoptions(Dialog) :

    def cb_setheader( self ) :
        value              = self.cb_header.get()
        if value == 1 :
            self.cb_header.set(0)
        else :
            self.cb_header.set(1)
        return

    def cb_setfilesize( self ) :
        value              = self.cb_filesize.get()
        if value == 1 :
            self.cb_filesize.set(0)
        else :
            self.cb_filesize.set(1)
        return

    def cb_sethierarchy( self ) :
        value              = self.cb_hierarchy.get()
        if value == 1 :
            self.cb_hierarchy.set(0)
        else :
            self.cb_hierarchy.set(1)
        return

    def cb_settimeticks( self ) :
        value              = self.cb_timeticks.get()
        if value == 1 :
            self.cb_timeticks.set(0)
        else :
            self.cb_timeticks.set(1)
        return

    def cb_setbones( self ) :
        value              = self.cb_bones.get()
        if value == 1 :
            self.cb_bones.set(0)
        else :
            self.cb_bones.set(1)
        return

    def cb_setalldata( self ) :
        value              = self.cb_alldata.get()
        if value == 1 :
            self.cb_alldata.set(0)
        else :
            self.cb_alldata.set(1)
        return

    def cb_setfooter( self ) :
        value              = self.cb_footer.get()
        if value == 1 :
            self.cb_footer.set(0)
        else :
            self.cb_footer.set(1)
        return

    def body( self, master ) :
        self.title( "Select Data for Summary" )

        # Variables to hold the check button states.
        self.cb_header     = IntVar()
        self.cb_filesize   = IntVar()  
        self.cb_hierarchy  = IntVar()  
        self.cb_timeticks  = IntVar()  
        self.cb_bones      = IntVar()  
        self.cb_alldata    = IntVar()  
        self.cb_footer     = IntVar()  

        # Check buttons.
        self.cb_head       = Checkbutton( master, text = "Header Data",                  state = NORMAL, variable = self.cb_header    , command = self.cb_setheader    )
        self.cb_file       = Checkbutton( master, text = "File Size sans Header/Footer", state = NORMAL, variable = self.cb_filesize  , command = self.cb_setfilesize  )
        self.cb_hier       = Checkbutton( master, text = "Bone Hierarchy Data",          state = NORMAL, variable = self.cb_hierarchy , command = self.cb_sethierarchy )
        self.cb_time       = Checkbutton( master, text = "Time Ticks Data",              state = NORMAL, variable = self.cb_timeticks , command = self.cb_settimeticks )
        self.cb_bone       = Checkbutton( master, text = "Bone Data",                    state = NORMAL, variable = self.cb_bones     , command = self.cb_setbones     )
        self.cb_data       = Checkbutton( master, text = "Quat, Anim, and Pose Data",    state = NORMAL, variable = self.cb_alldata   , command = self.cb_setalldata   )
        self.cb_foot       = Checkbutton( master, text = "Footer Data",                  state = NORMAL, variable = self.cb_footer    , command = self.cb_setfooter    )

        # Defaults are header and footer data for the summary.
        self.cb_head.select() 
        self.cb_foot.select() 
        self.cb_header.set(1)
        self.cb_footer.set(1)

        # Layout the check buttons.
        self.cb_head.grid( row = 0, column =0, sticky = W ) 
        self.cb_file.grid( row = 1, column =0, sticky = W ) 
        self.cb_hier.grid( row = 2, column =0, sticky = W ) 
        self.cb_time.grid( row = 3, column =0, sticky = W ) 
        self.cb_bone.grid( row = 4, column =0, sticky = W ) 
        self.cb_data.grid( row = 5, column =0, sticky = W ) 
        self.cb_foot.grid( row = 6, column =0, sticky = W ) 

        return

    def apply(self) :
        headflag           = self.cb_header.get()      
        fileflag           = self.cb_filesize.get()         
        hierflag           = self.cb_hierarchy.get()        
        timeflag           = self.cb_timeticks.get()        
        boneflag           = self.cb_bones.get()          
        dataflag           = self.cb_alldata.get()        
        footflag           = self.cb_footer.get()        

        # Convert to ints and hold in self structure.
        self.head          = int( headflag )           
        self.file          = int( fileflag )
        self.hier          = int( hierflag )        
        self.time          = int( timeflag )        
        self.bone          = int( boneflag )        
        self.data          = int( dataflag )        
        self.foot          = int( footflag )        
        return 

    def getvalues(self) :
        # Get an instance of the class dataflags().
        flags              = dataflags()

        # Set the fields.
        flags.header       = int( self.cb_header.get()    )  
        flags.filesize     = int( self.cb_filesize.get()  )  
        flags.hierarchy    = int( self.cb_hierarchy.get() )  
        flags.timeticks    = int( self.cb_timeticks.get() )  
        flags.bones        = int( self.cb_bones.get()     )  
        flags.alldata      = int( self.cb_alldata.get()   )    
        flags.footer       = int( self.cb_footer.get()    )    
        flags.isdir        = 1

        return flags  

# -----------------------------------------------------------------------------------
#                            isfileindescrskeleton()                                |
# -----------------------------------------------------------------------------------
def isfileindescrskeleton( fncas ):

    answer                 = False

    [ pcas, fcas, ecas ]   = splitpath( fncas )       
    shortname              = fcas + ecas
    shortname              = shortname.lower()

    fnskel                 = 'descr_skeleton.txt'
    fidskel                = open( fnskel, 'r' )
    for line in fidskel :
        lowcaseline        = line.lower()
        nch                = lowcaseline.find( shortname )
        if nch > -1 :
            answer         = True
            fidskel.close()
            return answer

    fidskel.close()
    return answer

# -----------------------------------------------------------------------------------
#                            surveycasdirectory()                                   |
# -----------------------------------------------------------------------------------
#                                                                                   |
#    This application prompts the user for a series of check boxes to select what   |
# survey data he wants from the .cas files.  It then shows a filechooser so the     |
# user can select ANY .cas file in that directory to indicate which directory he    |
# wants to survey.  The output goes into the current directory with filename        |
# consisting of the directory with _summary appended and file extension .txt.       |
#                                                                                   |
# -----------------------------------------------------------------------------------
def surveycasdirectory() :

    # Check buttons for data desired in summary.
    root = Tk()
    root.withdraw()
    dialog                 = getdataoptions(root)
    flags                  = dialog.getvalues()
    print 'in animationutilities flags.header    = ' + str( flags.header )
    print 'in animationutilities flags.filesize  = ' + str( flags.filesize )
    print 'in animationutilities flags.hierarchy = ' + str( flags.hierarchy )
    print 'in animationutilities flags.timeticks = ' + str( flags.timeticks )
    print 'in animationutilities flags.bones     = ' + str( flags.bones ) 
    print 'in animationutilities flags.alldata   = ' + str( flags.alldata )
    print 'in animationutilities flags.footer    = ' + str( flags.footer )
    print 'in animationutilities flags.isdir     = ' + str( flags.isdir )

    # Get filename from a filechooser.
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfncas              = d.go(os.curdir, "*.cas")
    if fullfncas == None :
        exit()

    # Split full filename into its three parts.
    [ pdir, fdir, edir ]   = splitpath( fullfncas )

    # Need to enumerate all the .cas files in the directory pointed at by the path pdir.
    namelist               = os.listdir( pdir )
    nfiles                 = len( namelist )

    # Output summary name.
    nlast                  = pdir.rfind( '\\' )
    fntxt                  = pdir[nlast+1:] + '_summary.txt'

    # --------------------------------------------
    # Iterate over the namelist and convert the  |
    # cas files to txt files.                    |
    # --------------------------------------------
    num_casfiles           = 0

    for ifile in namelist :
        BASEPOSEFLAG       = False
        FLAILINGFLAG       = False
        [ pcas, fcas, ecas ] = splitpath( ifile )       

        nbasepose          = fcas.find( 'basepose' )
        if nbasepose > -1:
            BASEPOSEFLAG   = True
            print 'We have the basepose file....'

        nflailingname  = fcas.find( 'die_flailing' )
        if nflailingname > -1 :
            FLAILINGFLAG   = True
            print 'We have the die_flailing file....'

        if ecas.lower() == '.cas' :
            print 'Converting file = ' + fcas + ecas
            fnin           = pdir + '\\' + fcas + ecas
            fnout          = pdir + '\\' + fcas + '.txt'

            # Get data of original cas file and write .txt version on fnout
            num_casfiles   = num_casfiles + 1
            isinskel       = isfileindescrskeleton( fnin )
            fidtxt         = open( fntxt, 'a' )
            fidtxt.write( '\n' )
            fname          = fcas + ecas
            if isinskel == True :
                fidtxt.write( fname.ljust(50) + ' TRUE  TRUE  TRUE     IS in descr_skeleton.txt\n' )
            else:
                fidtxt.write( fname.ljust(50) + ' FALSE FALSE FALSE is NOT in descr_skeleton.txt\n' )
            fidtxt.close()
            casfiledatalist = readcasfile( fnin, fntxt, flags )

    fidtxt                 = open( fntxt, 'a' )
    fidtxt.write( '\n' )
    fidtxt.write( 'Processed ' + str( num_casfiles ) + ' in this directory...\n' )
    fidtxt.close()
    print 'Completed converting ' + str( num_casfiles ) + ' files in directory ' + pdir

    return

# -----------------------------------------------------------------------------------
#                            gettxtheaderdata()                                     |
# -----------------------------------------------------------------------------------
def gettxtheaderdata( headerdatastring ) :

    # Return value.
    headerdata             = []

    # Tokenize the line.
    tokens                 = headerdatastring.split()

    # Allocate byte arrays for signature data.
    signaturebytetriple1   = array.array( 'b' )
    signaturebytetriple2   = array.array( 'b' )

    # 42 bytes of header here.
    float_version          = float(   tokens[0] )          #  4 bytes.
    int_thirtyeight        = int(     tokens[1] )          #  8 bytes.
    int_nine               = int(     tokens[2] )          # 12 bytes.
    int_zero1              = int(     tokens[3] )          # 16 bytes. We throw away the stored animtime. If user added frames it's wrong now.
    float_animtime         = float(   tokens[4] )          # 20 bytes. 
    int_one1               = int(     tokens[5] )          # 24 bytes. 
    int_zero2              = int(     tokens[6] )          # 28 bytes. 
    signaturebytetriple1.append( int( tokens[7] ) )        # 29 bytes. 
    signaturebytetriple1.append( int( tokens[8] ) )        # 30 bytes. 
    signaturebytetriple1.append( int( tokens[9] ) )        # 31 bytes. 
    int_one2               = int(     tokens[10] )         # 35 bytes. 
    int_zero3              = int(     tokens[11] )         # 39 bytes. 
    signaturebytetriple2.append( int( tokens[12] ) )       # 40 bytes. 
    signaturebytetriple2.append( int( tokens[13] ) )       # 41 bytes. 
    signaturebytetriple2.append( int( tokens[14] ) )       # 42 bytes. 

    # Form into list.
    headerdata.append( float_version )                     # Index 0.
    headerdata.append( int_thirtyeight )                   # Index 1. 
    headerdata.append( int_nine )                          # Index 2. 
    headerdata.append( int_zero1 )                         # Index 3. 
    headerdata.append( float_animtime )                    # Index 4. 
    headerdata.append( int_one1 )                          # Index 5. 
    headerdata.append( int_zero2 )                         # Index 6. 
    headerdata.append( signaturebytetriple1 )              # Index 7. 
    headerdata.append( int_one2 )                          # Index 8. 
    headerdata.append( int_zero3 )                         # Index 9. 
    headerdata.append( signaturebytetriple2 )              # Index 10. 

    return headerdata

# -----------------------------------------------------------------------------------
#                            gettxthierarchytree()                                  |
# -----------------------------------------------------------------------------------
def gettxthierarchytree( hierarchydatastring ) :

    # Return value.
    hierarchydata          = array.array( 'I' )            # Unsigned int array.

    # Tokenize the line.
    tokens                 = hierarchydatastring.split()

    # Append the tokens.
    ntokens                = len( tokens )
    for ii in range( 1, ntokens ) :
        hierarchydata.append( int( tokens[ii] ) )

    return hierarchydata

# -----------------------------------------------------------------------------------
#                            gettxttimeticks()                                      |
# -----------------------------------------------------------------------------------
def gettxttimeticksdata( timeticksdatastring ) :

    tokens                 = timeticksdatastring.split()
    ntokens                = len( tokens )

    timeticksdata          = array.array( 'f' )            # Float array.
    for ii in range( 1, ntokens ) :
        timeticksdata.append( float( tokens[ii] ) )

    return timeticksdata

# -----------------------------------------------------------------------------------
#                            gettxtbonesection()                                    |
# -----------------------------------------------------------------------------------
def gettxtbonesection( fidtxt, nbones ) :

    # This is the return value.
    bonedata               = []

    # Allocate arrays.
    bonenames              = []
    quatframesperbone      = array.array( 'I' )            # Unsigned int array.
    animframesperbone      = array.array( 'I' )            # Unsigned int array.
    quatoffsetperbone      = array.array( 'I' )            # Unsigned int array.
    animoffsetperbone      = array.array( 'I' )            # Unsigned int array.
    zerosperbone           = array.array( 'I' )            # Unsigned int array.
    onesperbone            = array.array( 'I' )            # Unsigned int array.

    # Process the text file to fill in lists.
    for ii in range( nbones ) :
        line               = fidtxt.readline()
        tokens             = line.split()
        if tokens[1] == 'Root' :
            bonenames.append( tokens[0] + ' ' + tokens[1] )
            quatframesperbone.append( int( tokens[2] ) )                 
            animframesperbone.append( int( tokens[3] ) )                    
            quatoffsetperbone.append( int( tokens[4] ) )
            animoffsetperbone.append( int( tokens[5] ) )
            zerosperbone.append(      int( tokens[6] ) )
            onesperbone.append(       int( tokens[7] ) )
        else :
            bonenames.append( tokens[0] )
            quatframesperbone.append( int( tokens[1] ) )                 
            animframesperbone.append( int( tokens[2] ) )                    
            quatoffsetperbone.append( int( tokens[3] ) )
            animoffsetperbone.append( int( tokens[4] ) )
            zerosperbone.append(      int( tokens[5] ) )
            onesperbone.append(       int( tokens[6] ) )

    # Form the data structure.
    bonedata.append( bonenames )
    bonedata.append( quatframesperbone )
    bonedata.append( animframesperbone )
    bonedata.append( quatoffsetperbone )
    bonedata.append( animoffsetperbone )
    bonedata.append( zerosperbone )
    bonedata.append( onesperbone )

    return bonedata

# -----------------------------------------------------------------------------------
#                            geteulerdata()                                         |
# -----------------------------------------------------------------------------------
def geteulerdata( fidtxt, bonedata ) :

    # This is the return value.
    eulers                 = array.array( 'f' )

    # Unpack the quatframesperbone.
    quatframesperbone      = bonedata[1]

    nbones                 = len( quatframesperbone ) - 1
    for ibone in range( nbones ) :
        nframes            = quatframesperbone[ibone+1]    # We use +1 to skip Scene_Root.
        if nframes > 0 :
            dummy          = fidtxt.readline()             # This is just the text separator line. 
            for iframe in range( nframes ) :
                line       = fidtxt.readline()
                tokens     = line.split()
                eulers.append( float( tokens[4] ) )
                eulers.append( float( tokens[5] ) )
                eulers.append( float( tokens[6] ) )

    return eulers

# -----------------------------------------------------------------------------------
#                            getanimdata()                                          |
# -----------------------------------------------------------------------------------
def getanimdata( fidtxt, bonedata ) :

    # This is the return value.
    animfloats             = array.array( 'f' )

    # Unpack the animframesperbone.
    animframesperbone      = bonedata[2]

    nbones                 = len( animframesperbone ) - 1
    for ibone in range( nbones ) :
        nframes            = animframesperbone[ibone+1]    # We use +1 to skip Scene_Root.
        if nframes > 0 :
            dummy          = fidtxt.readline()             # This is just the text separator line. 
            for iframe in range( nframes ) :
                line       = fidtxt.readline()
                tokens     = line.split()
                animfloats.append( float( tokens[0] ) )
                animfloats.append( float( tokens[1] ) )
                animfloats.append( float( tokens[2] ) )

    return animfloats

# -----------------------------------------------------------------------------------
#                            getposedata()                                          |
# -----------------------------------------------------------------------------------
def getposedata( fidtxt, bonedata ) :

    # This is the return value.
    posefloats             = array.array( 'f' )

    # Unpack the animframesperbone.
    animframesperbone      = bonedata[2]

    nbones                 = len( animframesperbone )       
    dummy                  = fidtxt.readline()             # Throw away the separator line.
    for ibone in range( nbones ) :
        line               = fidtxt.readline()
        tokens             = line.split()
        posefloats.append( float( tokens[0] ) )
        posefloats.append( float( tokens[1] ) )
        posefloats.append( float( tokens[2] ) )

    return posefloats

# -----------------------------------------------------------------------------------
#    Processes .txt footer string.  There are four footer types we have to contend  |
# with.                                                                             |
#                                                                                   |
# (1) Regular footer of 192 bytes.  This is our standard footer with regular        |
#     fields.                                                                       |
#                                                                                   |
# (2) Short footer by 12 bytes.  The final three ints, 12 12 0, are missing so the  |
#     footer is 180 bytes long.  This is fairly common.  Just make int_vec2 into    |
#     an empty list and take care not to try to write it out.                       |
#                                                                                   |
# (3) Camel footer.  Very short, we'll test and process for it first.               |
#                                                                                   |
# (4) Long footer.  This occurs with fs_horse directory .cas files.  Indicated by   |
#     int_104 actually being equal to 170.  There's a lot of extra float fields to  |
#     contend with here so the footerdata list gets more complicated.               |
#                                                                                   |
# -----------------------------------------------------------------------------------
def gettxtfooterdata( footerdatastring ) :

    # HORSEFLAG is used for doing long footers in fs_horse.
    HORSEFLAG              = False
    CAMELFLAG              = False

    # This is the return value.
    footerdata             = []

    # Tokenize the string.
    tokens                 = footerdatastring.split()

    # Read the first token to see how to process the rest.
    int_104                = int( tokens[0] )         

    # Check to see if int_104 is equal to 170.  If so, we are doing a long fs_horse footer
    # with lots of extra float data.
    if int_104 == 170 :
        HORSEFLAG          = True

    # Check to see if int_104 is equal to 18.  If so, we are doing a very short camel footer.
    if int_104 == 18 :
        CAMELFLAG          = True

    # If CAMELFLAG is True, process it now because it's short and we can return it and be done.
    if CAMELFLAG == True :
        int_one1           = int( tokens[1] )     
        int_zero1          = int( tokens[2] ) 
        int_zero2          = int( tokens[3] ) 
        short_zero1        = int( tokens[4] ) 

        int_vec1           = array.array( 'I' )
        int_vec1.append( int( tokens[5] ) )
        int_vec1.append( int( tokens[6] ) )
        int_vec1.append( int( tokens[7] ) )
        int_vec1.append( int( tokens[8] ) )

        int_vec2           = array.array( 'I' )
        int_vec2.append( int( tokens[9] ) )
        int_vec2.append( int( tokens[10] ) )
        int_vec2.append( int( tokens[11] ) )
        int_vec2.append( int( tokens[12] ) )

        int_vec3           = array.array( 'I' )
        int_vec3.append( int( tokens[13] ) )
        int_vec3.append( int( tokens[14] ) )
        int_vec3.append( int( tokens[15] ) )
        int_vec3.append( int( tokens[16] ) )

        int_vec4           = array.array( 'I' )
        int_vec4.append( int( tokens[17] ) )
        int_vec4.append( int( tokens[18] ) )
        int_vec4.append( int( tokens[19] ) )
        int_vec4.append( int( tokens[20] ) )

        int_vec5           = array.array( 'I' )
        int_vec5.append( int( tokens[21] ) )
        int_vec5.append( int( tokens[22] ) )
        int_vec5.append( int( tokens[23] ) )

        footerdata.append( int_104 )
        footerdata.append( int_one1 )
        footerdata.append( int_zero1 )
        footerdata.append( int_zero2 )
        footerdata.append( short_zero1 )
        footerdata.append( int_vec1 )
        footerdata.append( int_vec2 )
        footerdata.append( int_vec3 )
        footerdata.append( int_vec4 )
        footerdata.append( int_vec5 )
        
        return footerdata

    # Continue regular processing.
    int_one1               = int( tokens[1] )   
    int_one2               = int( tokens[2] ) 
    AttribString           = tokens[3]


    # Now read an int and a byte, then two ints and a byte.
    int_one3               = int( tokens[4] ) 
    byte_zero1             = int( tokens[5] ) 
    int_one4               = int( tokens[6] ) 
    int_zero1              = int( tokens[7] ) 
    byte_zero2             = int( tokens[8] ) 

    # Now read 7 floats.
    float_vec1             = array.array( 'f' )
    float_vec1.append( float( tokens[9] ) )
    float_vec1.append( float( tokens[10] ) )
    float_vec1.append( float( tokens[11] ) )
    float_vec1.append( float( tokens[12] ) )
    float_vec1.append( float( tokens[13] ) )
    float_vec1.append( float( tokens[14] ) )
    float_vec1.append( float( tokens[15] ) )

    # Now read an int and a short.
    int_zero2              = int( tokens[16] ) 
    short_zero1            = int( tokens[17] ) 

    # Now read a -1 int and another int zero.
    int_minus1             = int( tokens[18] ) 
    int_zero3              = int( tokens[19] ) 

    # Save off the common data.
    footerdata.append( int_104 )                           # Index 0.
    footerdata.append( int_one1 )                          # Index 1.
    footerdata.append( int_one2 )                          # Index 2.
    footerdata.append( AttribString )                      # Index 3. (Remember the null terminator byte.)
    footerdata.append( int_one3 )                          # Index 4.
    footerdata.append( byte_zero1 )                        # Index 5.
    footerdata.append( int_one4 )                          # Index 6.
    footerdata.append( int_zero1 )                         # Index 7.
    footerdata.append( byte_zero2 )                        # Index 8.
                                             
    footerdata.append( float_vec1 )                        # Index 9.
                                             
    footerdata.append( int_zero2 )                         # Index 10.
    footerdata.append( short_zero1 )                       # Index 11.
    footerdata.append( int_minus1 )                        # Index 12.
    footerdata.append( int_zero3 )                         # Index 13.

    # Must split processing depending on if its a fs_horse file or not.
    if HORSEFLAG == False :
        short_zero2        = int( tokens[20] ) 
        int_something      = int( tokens[21] ) 

        # Now read a long bit of 19 ints.
        int_vec1           = array.array( 'I' )
        int_vec1.append( int( tokens[22] ) )
        int_vec1.append( int( tokens[23] ) )
        int_vec1.append( int( tokens[24] ) )
        int_vec1.append( int( tokens[25] ) )

        int_vec2           = array.array( 'I' )
        int_vec2.append( int( tokens[26] ) )
        int_vec2.append( int( tokens[27] ) )
        int_vec2.append( int( tokens[28] ) )
        int_vec2.append( int( tokens[29] ) )

        int_vec3           = array.array( 'I' )
        int_vec3.append( int( tokens[30] ) )
        int_vec3.append( int( tokens[31] ) )
        int_vec3.append( int( tokens[32] ) )
        int_vec3.append( int( tokens[33] ) )

        int_vec4           = array.array( 'I' )
        int_vec4.append( int( tokens[34] ) )
        int_vec4.append( int( tokens[35] ) )
        int_vec4.append( int( tokens[36] ) )
        int_vec4.append( int( tokens[37] ) )

        int_vec5           = array.array( 'I' )
        int_vec5.append( int( tokens[38] ) )
        int_vec5.append( int( tokens[39] ) )
        int_vec5.append( int( tokens[40] ) )

        # MAYBE read 3 ints if not a bad file.
        if len( tokens ) == 41 :
            int_vec6       = []
        else :
            int_vec6       = array.array( 'I' )
            int_vec6.append( int( tokens[41] ) )
            int_vec6.append( int( tokens[42] ) )
            int_vec6.append( int( tokens[43] ) )

        # Save off the data pertaining to non fs_horse.
        footerdata.append( short_zero2 )                   # Index 14.
        footerdata.append( int_something )                 # Index 15.
        footerdata.append( int_vec1 )                      # Index 16.
        footerdata.append( int_vec2 )                      # Index 17.
        footerdata.append( int_vec3 )                      # Index 18.
        footerdata.append( int_vec4 )                      # Index 19.
        footerdata.append( int_vec5 )                      # Index 20.
        footerdata.append( int_vec6 )                      # Index 21.
    else :
        # Doing the horsies. 
        string2            = tokens[20]
        int_one5           = int( tokens[21] )
        byte_zero3         = int( tokens[22] )
        int_one6           = int( tokens[23] )
        int_zero4          = int( tokens[24] )
        byte_zero4         = int( tokens[25] )

        float_vec2         = array.array( 'f' )
        float_vec2.append( float( tokens[26] ) )
        float_vec2.append( float( tokens[27] ) )
        float_vec2.append( float( tokens[28] ) )
        float_vec2.append( float( tokens[29] ) )
        float_vec2.append( float( tokens[30] ) )
        float_vec2.append( float( tokens[31] ) )
        float_vec2.append( float( tokens[32] ) )

        int_vec1           = array.array( 'I' )
        int_vec1.append( int( tokens[33] ) )
        int_vec1.append( int( tokens[34] ) )
        int_vec1.append( int( tokens[35] ) )
        int_vec1.append( int( tokens[36] ) )
        int_vec1.append( int( tokens[37] ) )

        int_vec2           = array.array( 'I' )
        int_vec2.append( int( tokens[38] ) )
        int_vec2.append( int( tokens[39] ) )
        int_vec2.append( int( tokens[40] ) )
        int_vec2.append( int( tokens[41] ) )

        int_vec3           = array.array( 'I' )
        int_vec3.append( int( tokens[42] ) )
        int_vec3.append( int( tokens[43] ) )
        int_vec3.append( int( tokens[44] ) )
        int_vec3.append( int( tokens[45] ) )

        int_vec4           = array.array( 'I' )
        int_vec4.append( int( tokens[46] ) )
        int_vec4.append( int( tokens[47] ) )
        int_vec4.append( int( tokens[48] ) )
        int_vec4.append( int( tokens[49] ) )

        int_vec5           = array.array( 'I' )
        int_vec5.append( int( tokens[50] ) )
        int_vec5.append( int( tokens[51] ) )
        int_vec5.append( int( tokens[52] ) )
        int_vec5.append( int( tokens[53] ) )

        int_count          = int( tokens[54] ) 
        int_five           = int( tokens[55] ) 
        int_one7           = int( tokens[56] ) 
        int_zero5          = int( tokens[57] ) 

        float_vec3         = array.array( 'f' )
        float_vec3.append( float( tokens[58] ) )
        float_vec3.append( float( tokens[59] ) )
        float_vec3.append( float( tokens[60] ) )
        float_vec3.append( float( tokens[61] ) )
        float_vec3.append( float( tokens[62] ) )
        float_vec3.append( float( tokens[63] ) )
        float_vec3.append( float( tokens[64] ) )

        int_vec6           = array.array( 'I' )
        int_vec6.append( int( tokens[65] ) )
        int_vec6.append( int( tokens[66] ) )
        int_vec6.append( int( tokens[67] ) )
        int_vec6.append( int( tokens[68] ) )
        int_vec6.append( int( tokens[69] ) )
        int_vec6.append( int( tokens[70] ) )
        int_vec6.append( int( tokens[71] ) )

        byte_zero5         = int( tokens[72] ) 

        int_vec7           = array.array( 'I' )
        int_vec7.append( int( tokens[73] ) )
        int_vec7.append( int( tokens[74] ) )
        int_vec7.append( int( tokens[75] ) )

        # Save off the data.
        footerdata.append( string2 )                       # Index 14.
        footerdata.append( int_one5 )                      # Index 15.
        footerdata.append( byte_zero3 )                    # Index 16.
        footerdata.append( int_one6 )                      # Index 17.
        footerdata.append( int_zero4 )                     # Index 18.
        footerdata.append( byte_zero4 )                    # Index 19.
        footerdata.append( float_vec2 )                    # Index 20.
        footerdata.append( int_vec1 )                      # Index 21.
        footerdata.append( int_vec2 )                      # Index 22.
        footerdata.append( int_vec3 )                      # Index 23.
        footerdata.append( int_vec4 )                      # Index 24.
        footerdata.append( int_vec5 )                      # Index 25.
        footerdata.append( int_count )                     # Index 26.
        footerdata.append( int_five )                      # Index 27.
        footerdata.append( int_one7 )                      # Index 28.
        footerdata.append( int_zero5 )                     # Index 29.
        footerdata.append( float_vec3 )                    # Index 30.
        footerdata.append( int_vec6 )                      # Index 31.
        footerdata.append( byte_zero5 )                    # Index 32.
        footerdata.append( int_vec7 )                      # Index 33.

        return footerdata

    return footerdata


# -----------------------------------------------------------------------------------
#                            converttxttocas()                                      |
# -----------------------------------------------------------------------------------
#                                                                                   |
#    This application prompts the user for a .txt file to convert back to a binary  |
# .cas file.  The assumption made here is that the user may have modified the       |
# animation data and deltas, the poses, and the rotation data by changing the euler |
# angles and NOT the quaternions. (Who thinks in quaternions anyway.)  So the       |
# quaternions are ignored and the euler angles are read and reprocessed into the    |
# correct quaternions to put into the .cas file.                                    |
#                                                                                   |
# -----------------------------------------------------------------------------------
def converttxttocas() :

    # Get filename from a filechooser.
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfntxt              = d.go(os.curdir, "*.txt")
    if fullfntxt == None :
        exit()

    # Split full filename into its three parts.
    [ pdir, fdir, edir ]   = splitpath( fullfntxt )

    # Make an output .cas filename.
    fullfncas              = pdir + '\\' + fdir + '_modified.cas'

    # ------------------------------------------------------
    # Accumulate data from the .txt file.                  |
    # ------------------------------------------------------

    # Open the .txt as an ASCII file and start reading.
    fidtxt                 = open( fullfntxt, 'r' )

    # Header.
    headerdatastring       = fidtxt.readline()
    headerdata             = gettxtheaderdata( headerdatastring ) 

    # Filesize sans header/footer.
    filesizestring         = fidtxt.readline()
    tokens                 = filesizestring.split()
    filesizesans           = int( tokens[0] )
    zeroint                = int( tokens[1] )

    # Hierarchy.
    hierarchydatastring    = fidtxt.readline()
    hierarchydata          = gettxthierarchytree( hierarchydatastring ) 
    nbones                 = len( hierarchydata )

    # Time ticks.
    timeticksdatastring    = fidtxt.readline()
    timeticksdata          = gettxttimeticksdata( timeticksdatastring ) 

    # Bone data.
    bonedata               = gettxtbonesection( fidtxt, nbones ) 

    # Process the euler angles and get quaternions.
    eulers                 = geteulerdata( fidtxt, bonedata )
    quatfloats             = computequats( eulers )

    # Process the anim data.
    animfloats             = getanimdata( fidtxt, bonedata )

    # Process the pose data.
    posefloats             = getposedata( fidtxt, bonedata )

    # Process the footer.
    footerdatastring       = fidtxt.readline()
    if footerdatastring == '' :
        footerdata         = []
    else :
        footerdata         = gettxtfooterdata( footerdatastring )

    # ------------------------------------------------------
    # Store in casfiledatalist.                            |
    # ------------------------------------------------------
    casfiledatalist        = []
    casfiledatalist.append( headerdata )
    casfiledatalist.append( filesizesans )
    casfiledatalist.append( zeroint )     
    casfiledatalist.append( hierarchydata )
    casfiledatalist.append( timeticksdata )
    casfiledatalist.append( bonedata )        
    casfiledatalist.append( quatfloats )      
    casfiledatalist.append( animfloats )      
    casfiledatalist.append( posefloats )      
    casfiledatalist.append( eulers )          
    casfiledatalist.append( footerdata )      

    # Use the library function.
    writecasfile( fullfncas, casfiledatalist )

    return


# -----------------------------------------------------------------------------------
#                            exportextrabonesskeleton()                             |
# -----------------------------------------------------------------------------------
#                                                                                   |
#    This application exports a skeleton with extra bones and hierarchy data to     |
# a directory of .cas files.  The user selects a skeleton file he has prepared with |
# a GUI filechooser, then selects the directory to export to by mousing to it in    |
# another filechooser and selecting any file in the directory, and finally the      |
# user can choose to scale the x, y, and z animation data in case his skeleton is   |
# of non-standard size.                                                             |
#                                                                                   |
#    The extra bone quat data are just frames of 0.0, 0.0, 0.0, 1.0 no rotation     |
# quats and the extra anim delta data are just frames of 0.0, 0.0, 0.0 deltas.      |
# The extra bone data is added at the end of the posedata.                          |
#                                                                                   |
#    The bone section is fixed up to have the new bone included using the number of |
# bone_pelvis quat frames and anim frames. (This handles the basepose files which   |
# don't have this data so nothing is added except the hierarchy tree, the extra     |
# bone name in the bone section, and the extra bone in the pose data.)              |
#                                                                                   |
# -----------------------------------------------------------------------------------
def exportextrabonesskeleton() :

    # --------------------------------------------
    # Construct a flags object to pass.          |
    # --------------------------------------------
    flags                  = dataflags()
    flags.header           = 0
    flags.filesize         = 0
    flags.hierarchy        = 0
    flags.timeticks        = 0
    flags.bones            = 0
    flags.alldata          = 0
    flags.footer           = 0
    flags.isdir            = 0

    # --------------------------------------------
    # Get a skeleton filename via a GUI.         |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfnskeleton         = d.go(os.curdir, "*.skeleton")
    if fullfnskeleton == None :
        exit()

    # --------------------------------------------
    # Get any cas filename in directory to be    |
    # processed.                                 |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfncas              = d.go(os.curdir, "*.cas")
    if fullfncas == None :
        exit()

    # Split full filename into its three parts.
    [ pdir, fdir, edir ]   = splitpath( fullfncas )

    # --------------------------------------------
    # Get the scaling factors from the user.     |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    dialog                 = getscalingfactors(root)
    factors                = dialog.getvalues()

    SCALEX                 = factors[0] 
    SCALEY                 = factors[1] 
    SCALEZ                 = factors[2] 
    print 'SCALEX, Y, Z = ' + str( SCALEX ) + '  ' + str( SCALEY ) + ' ' + str( SCALEZ )

    # --------------------------------------------
    # Read the skeleton data into newposefloats. |
    # Also read the new hierarchy tree data      |
    # since this has extra bones now.            |
    # --------------------------------------------
    fidskel                = open( fullfnskeleton, 'r' )
    newposefloats          = array.array( 'f' )
    newposefloats.append( 0.0 )                            # These are Scene_Root poses.
    newposefloats.append( 0.0 )
    newposefloats.append( 0.0 )
    newhierarchydata       = array.array( 'I' )
    newhierarchydata.append( 0 )                           # This is Scene_Root hierarchy entry.
    newbonenames           = []
    while 1 == 1 :
        line               = fidskel.readline()
        tokens             = line.split()
        if len( tokens ) > 1 :
            newposefloats.append(  float( tokens[0] ) )
            newposefloats.append(  float( tokens[1] ) )
            newposefloats.append(  float( tokens[2] ) )
            newhierarchydata.append( int( tokens[3] ) )
            newbonenames.append(          tokens[4] )      
        else :
            break

    fidskel.close()

    # Need to enumerate all the .cas files in the directory pointed at by the path pdir.
#    namelist               = [ 'c:\\sys\\downloads\\medievalII\\meshreader_h2w_611\\packextractor\\MTW2_Spear_charge.cas' ]
    namelist               = os.listdir( pdir )
    nfiles                 = len( namelist )

    # Create a subdirectory to hold the rewritten files.
    pconvdir               = pdir + '\\convertedfiles\\'
    direxists              = os.path.isdir( pconvdir )
    if direxists == False :
        os.makedirs( pconvdir )

    # --------------------------------------------
    # Iterate over the namelist and put the new  |
    # pose data over the old, scale the anim data|
    # and put in default quats and zero deltas   |
    # for the new bones, and redo the bonedata   |
    # section to account for the new bones as    |
    # well.                                      |
    # --------------------------------------------
    for ifile in namelist :
        BASEPOSEFLAG       = False
        FLAILINGFLAG       = False
        print 'processing file = ' + ifile
        [ pcas, fcas, ecas ] = splitpath( ifile )       

        nbasepose          = fcas.find( 'basepose' )
        if nbasepose > -1:
            BASEPOSEFLAG   = True
            print 'We have the basepose file....'

        nflailingname  = fcas.find( 'die_flailing' )
        if nflailingname > -1 :
            FLAILINGFLAG   = True
            print 'We have the die_flailing file....'

        if ecas == '.cas' :
            fnin           = pdir + '\\' + fcas + ecas
            fnout          = pconvdir    + fcas + ecas
            #print 'fnin  = ' + fnin
            #print 'fnout = ' + fnout

            # Get data of original cas file.
            casfiledatalist = readcasfile( fnin, [], flags )

            # Unpack the data that has to be modified.  
            # Hierarchy tree data is index 3, 
            # Bone data is index 5,
            # quatfloats is index 6,
            # animfloats is index 7,
            # posefloats is index 8.
            bonedata        = casfiledatalist[5]
            quatfloats      = casfiledatalist[6]
            animfloats      = casfiledatalist[7]

            # ----------------------------------------------
            # Substitute our new hierarchy tree.           |
            # ----------------------------------------------
            casfiledatalist[3] = newhierarchydata

            # ----------------------------------------------
            # Substitute our new skeleton.                 |
            # ----------------------------------------------
            casfiledatalist[8] = newposefloats             

            # ----------------------------------------------
            # Fix up the bone section data.                |
            # ----------------------------------------------
            bonenames              = bonedata[0]           # NOTE: These all include Scene_Root.
            quatframesperbone      = bonedata[1] 
            animframesperbone      = bonedata[2] 
            zerosperbone           = bonedata[5] 
            onesperbone            = bonedata[6] 

            nquatframes            = quatframesperbone[1]  # Use bone_pelvis data in case it's a basepose file. (No quats or anims.)
            nanimframes            = animframesperbone[1]
            print 'nquatframes = ' + str( nquatframes )
            print 'nanimframes = ' + str( nanimframes )

            noldbones              = len( bonenames )      # Careful! Scene_Root is a bone but isn't in the skeleton file.
            nnewbones              = len( newbonenames )   # Scene_Root is NOT included here.
            num_newbones           = nnewbones - noldbones + 1
            print 'noldbones    = ' + str( noldbones )
            print 'nnewbones    = ' + str( nnewbones )
            print 'num_newbones = ' + str( num_newbones )

            for ii in range( (noldbones-1), nnewbones ) :
                bonenames.append( newbonenames[ii] )
                quatframesperbone.append( nquatframes )
                animframesperbone.append( nanimframes )
                zerosperbone.append( 0 )
                onesperbone.append( 1 )

            # Now update the offsets.  First count the quat frames.
            totalquatframes  = 0
            for ii in range( len( quatframesperbone ) ) :
                totalquatframes = totalquatframes + quatframesperbone[ii]        

            # This is where the anim data will start.
            animsbaseoffset = totalquatframes * 16

            # Allocate the arrays for the offsets.
            quatoffsetperbone = array.array( 'I' )
            animoffsetperbone = array.array( 'I' )

            # Do the Scene_Root bone separately.
            quatoffsetperbone.append( 0 )
            animoffsetperbone.append( animsbaseoffset )

            nquatbytes     = 0
            nanimbytes     = animsbaseoffset
            for ii in range( nnewbones ) :
                quatoffsetperbone.append( nquatbytes )
                animoffsetperbone.append( nanimbytes )
                nquatbytes = nquatbytes + 16 * quatframesperbone[ii+1]
                nanimbytes = nanimbytes + 12 * animframesperbone[ii+1]

            print 'length of quatoffsetperbone = ' + str( len(quatoffsetperbone) )

            bonedata       = []
            bonedata.append( bonenames )
            bonedata.append( quatframesperbone )
            bonedata.append( animframesperbone )
            bonedata.append( quatoffsetperbone )
            bonedata.append( animoffsetperbone )
            bonedata.append( zerosperbone )
            bonedata.append( onesperbone )

            casfiledatalist[5] = bonedata

            # ----------------------------------------------
            # Add 0.0 0.0 0.0 1.0 quats for new bones.     |  NOTE: This section could be replaced by a stock
            # ----------------------------------------------        animation section for automating.
            for ibone in range( num_newbones ) :
                for iframe in range( nquatframes ) :
                    quatfloats.append( 0.0 )
                    quatfloats.append( 0.0 )
                    quatfloats.append( 0.0 )
                    quatfloats.append( 1.0 )

            casfiledatalist[6] = quatfloats

            # ----------------------------------------------
            # Add 0.0 0.0 0.0 anim deltas for new bones.   |
            # ----------------------------------------------
            for ibone in range( num_newbones ) :
                for iframe in range( nanimframes ) :
                    animfloats.append( 0.0 )
                    animfloats.append( 0.0 )
                    animfloats.append( 0.0 )

            # ----------------------------------------------
            # Actually SCALE the data this time you idiot! |
            # ----------------------------------------------
            nfloats            = len( animfloats )
            nvectors           = nfloats / 3
            for ii in range( nvectors ) :
                animfloats[3*ii+0] = SCALEX * animfloats[3*ii+0]
                animfloats[3*ii+1] = SCALEY * animfloats[3*ii+1]
                animfloats[3*ii+2] = SCALEZ * animfloats[3*ii+2]

            casfiledatalist[7] = animfloats


            # Debugging stuff.
#            print 'Last few added quatframes'
#            for ibone in range( nnewbones-3, nnewbones ):
#                print 'ibone = ' + str( ibone )
#                for iframe in range( nquatframes ) :
#                    idx    = ibone * nquatframes * 4 + iframe * 4
#                    print str( quatfloats[idx+0] ) + '   ' + str( quatfloats[idx+1] ) + '  ' + str( quatfloats[idx+2] ) + '   ' + str( quatfloats[idx+3] ) 
#
#            print 'Last few added animframes'
#            for ibone in range( nnewbones-3, nnewbones ):
#                print 'ibone = ' + str( ibone )
#                for iframe in range( nanimframes ) :
#                    idx    = ibone * nanimframes * 3 + iframe * 3
#                    print str( animfloats[idx+0] ) + '   ' + str( animfloats[idx+1] ) + '  ' + str( animfloats[idx+2] ) 
#
#            print 'All the poseframes'
#            for ibone in range( nnewbones+1 ):
#                idx        = ibone * 3
#                print str( newposefloats[idx+0] ) + '   ' + str( newposefloats[idx+1] ) + '  ' + str( newposefloats[idx+2] ) 
#
#            print 'The quatframesperbone and bonenames'
#            for ibone in range( nnewbones + 1 ) :
#                print 'ibone = ' + str(ibone) + ' quatframesperbone[ibone] = ' + str( quatframesperbone[ibone] )  + ', bonename = ' + bonenames[ibone]
#
#            print 'Length of bonenames = ' + str( len(bonenames) )


            # Write new cas file out to the \convertedfiles subdirectory.
            writecasfile( fnout, casfiledatalist )



    print '  '
    print 'exportextrabonesskeleton has completed processing all the .cas files in this directory'

    return


# -----------------------------------------------------------------------------------
#                            reorderbones()                                         |
# -----------------------------------------------------------------------------------
#                                                                                   |
#    This application reorders the bones of a .ms3d file that has extra bones after |
# the weapon/shield bones.  The user simply selects the .ms3d file from a           |
# filechooser and the utility outputs a new file in the same directory with the     |
# string "_reordered" appended to it.                                               |
#                                                                                   |
#    Reordering the joint data is easy but you also have to reorder the primary     |
# bone Id assignments in vertexdata and the secondary bone Id assignments in        |
# weightdata.                                                                       |
#                                                                                   |
# -----------------------------------------------------------------------------------
def reorderbones() :

    # --------------------------------------------
    # Get .ms3d filename to be reordered.        |
    # --------------------------------------------
    root = Tk()
    root.withdraw()
    d                      = FileDialog(root)
    fullfnms3d             = d.go(os.curdir, "*.ms3d")
    if fullfnms3d == None :
        exit()

    # Split full filename into its three parts.
    [ pdir, fdir, edir ]   = splitpath( fullfnms3d )

    # Construct output filename.
    fnout                  = pdir + "\\" + fdir + "_reordered.ms3d"

    # --------------------------------------------
    # Read .ms3d file.                           |
    # --------------------------------------------
    ms3dfiledatalist       = readms3dfile( fullfnms3d, [] )

    vertexdata             = ms3dfiledatalist[1]
    jointdata              = ms3dfiledatalist[6]
    weightdata             = ms3dfiledatalist[13]

    vbonesPrimary          = vertexdata[2]
    vbonesSecondary        = weightdata[0]
    nvert                  = len( vbonesPrimary )
    njoints                = len( jointdata )

    # --------------------------------------------
    # Extract bonenames.                         |
    # --------------------------------------------
    bonenames              = []
    attachmentindices      = []
    extraindices           = []
    HITWEAPONSFLAG         = False
    for ii in range( njoints ) :
        joint              = jointdata[ii]
        bonename           = joint[1]
        bonenames.append( bonename )
        nweapon            = bonename.find( 'weapon' )
        nshield            = bonename.find( 'shield' )
        if ( nweapon > -1 ) | ( nshield > - 1 ) :
            attachmentindices.append( ii )
            HITWEAPONSFLAG = True
        else :
            if HITWEAPONSFLAG == True :
                extraindices.append( ii )

    # Check for extra bones beyond weapon/shield bones.
    att_startidx           = attachmentindices[0]
    att_stopidx            = attachmentindices[len(attachmentindices)-1]
    if att_stopidx < njoints-1 :
        button             = askokcancel( "Proceed or Exit Dialog", "Extra bones found beyound the weapon/shield bones. Reorder them?", default = "cancel" )
        if button == False :
            return
    else :
        button             = showinfo( "Info", "No bones found beyond the weapon/shield bones.  No need to reorder" )
        return

    # --------------------------------------------
    # Reorder joints.                            |
    # --------------------------------------------
    att_joints             = []
    for ii in range( len( attachmentindices ) ) :
        joint              = jointdata[attachmentindices[ii]]
        att_joints.append( joint )
    extra_joints           = []
    for ii in range( len( extraindices ) ) :
        joint              = jointdata[extraindices[ii]]
        extra_joints.append( joint )

    # Move the extra bones down in index number.
    for ii in range( len( extraindices ) ) :
        joint              = extra_joints[ii]
        jointdata[extraindices[ii]-len(attachmentindices)] = joint # This moves it down.

    # Move the weapons/shield up in index number.
    for ii in range( len( attachmentindices ) ) :
        joint              = att_joints[ii]
        jointdata[attachmentindices[ii]+len(extraindices)] = joint # This moves it up.

    # Debug stuff.
    for ii in range( njoints ) :
        joint              = jointdata[ii]
        bonename           = joint[1]
        print "For index ii = " + str( ii ) + ", bonename = " + bonename

    # --------------------------------------------
    # Renumber primary bone IDs.                 |
    # --------------------------------------------
    for ii in range( len( extraindices ) ) :
        oldboneId          = extraindices[ii]
        newboneId          = extraindices[ii] - len( attachmentindices ) 
        for ivert in range( nvert ) :
            if vbonesPrimary[ivert] == oldboneId :
                vbonesPrimary[ivert] = newboneId

    vertexdata[2]          = vbonesPrimary

    # --------------------------------------------
    # Renumber secondary bone IDs.               |
    # --------------------------------------------
    for ii in range( len( extraindices ) ) :
        oldboneId          = extraindices[ii]
        newboneId          = extraindices[ii] - len( attachmentindices ) 
        for ivert in range( nvert ) :
            if vbonesSecondary[ivert] == oldboneId :
                vbonesSecondary[ivert] = newboneId

    weightdata[0]          = vbonesSecondary

    # --------------------------------------------
    # Reassign updated info.                     |
    # --------------------------------------------
    ms3dfiledatalist[1]    = vertexdata             
    ms3dfiledatalist[6]    = jointdata              
    ms3dfiledatalist[13]   = weightdata             

    # --------------------------------------------
    # Write the reordered .ms3d file.            |
    # --------------------------------------------
    print 'Writing the reordered file to ' + fnout
    writems3dfile( fnout, ms3dfiledatalist )

    return


# ===================================================================================
#                            Button callback functions.                             |
# ===================================================================================

def animmergeButton( event = None ) :
    animmerge()
    return

def animextractButton( event = None ) :
    animextract()
    return

def exportSkelButton( event = None ) :
    exportskeleton()
    return

def extractSkelButton( event = None ) :
    extractskeleton()
    return

def convertcasButton( event = None ) :
    convertcastotxt()
    return

def surveycasdirButton( event = None ) :
    surveycasdirectory()
    return

def converttxtButton( event = None ) :
    converttxttocas()
    return

def exportExtraBonesSkelButton( event = None ) :
    exportextrabonesskeleton()
    return

def reorderbonesButton( event = None ) :
    reorderbones()
    return


# ===================================================================================
#                            Info button help functions.                            |
# ===================================================================================

def InfoanimmergeButton( event = None ):

    infostring = """
       The animmerge utility is used to merge a Milkshape binary .ms3d file with
    one of the full format .cas animation files in Caliban's animations directory.       
    A filechooser will appear from which the user selects a .ms3d file.  A second 
    filechooser then appears from which the .cas file is selected.  Finally a 
    data entry dialog appears from which the user sets the Frames Per Second for the          
    animation. (20 is game speed, 2 to 5 is reasonably slow to study the animation.)  

       The output filename is the concatenation of the .ms3d filename with the .cas 
    filename, joined by the string '_animby_'.  For example, if the user merges 
    'armored_sergeants_lod0.ms3d' with the animation file 'MTW2_Spear_charge.cas',
    the resulting composite name would be 

                 'armored_sergeants_lod_animby_MTW2_Spear_charge.ms3d'.  

    This file will appear in the same directory as the input .ms3d file the user
    selected.

    Programming Note:  The .cas file header, hierarchy tree, and footer (if present)
    are stored as ASCII token strings in the first three joint comments.  Do not
    modify these comments or the animation is unlikely to be extracted out again
    properly.
    """

    showinfo( 'Animmerge Help', infostring )

    return


def InfoanimextractButton( event = None ):

    infostring = """
       The animextract utility is used after a user has used animmerge to create 
    an animated Milkshape .ms3d file and has modified the animation in some way
    in Milkshape.  A filechooser will appear from which the user selects the desired
    .ms3d file from which to extract the animation.  

       The filename of the extracted animation will be the original .cas filename 
    (which is part of the concatenated name following the string '_animby_') with 
    the string '_modified' added to it.  For example, if the user had merged 
    'armored_sergeants_lod0.ms3d' with the animation file 'MTW2_Spear_charge.cas',
    the resulting composite name would be 

                 'armored_sergeants_lod_animby_MTW2_Spear_charge.ms3d'.  

    The extracted animation after running animextract would then be 

                 'MTW2_Spear_charge_modified.cas'.
    
    This file will appear in the same directory as the animated .ms3d file the user
    selected.
    """

    showinfo( 'Animextract Help', infostring )

    return


def InfoextractSkelButton( event = None ):

    infostring = """
       The extractskeleton utility is used to extract the skeleton pose data from a 
    Milkshape .ms3d file.  This is intended as a convenience function for modelers
    making units with non-standard skeletons and who wish to export the skeleton
    to animations using the exportskeleton  or exportextrabonesskeleton utilities.  

       The name of the output file will be 'extracted.skeleton'.  The user should
    rename this to something more descriptive like 'dwarf.skeleton' or 'orc.skeleton'.
    This file will appear in the same directory as this utility.

    Programming Note:  The format of the skeleton file is x, y, and z bone offsets
    followed by the hierarchy tree entry and then the bone name.  Scene_Root is not
    included in this file.  Note that the sign of the x bone offset is opposite that
    of Milkshape.  This is because the skeleton is assumed to be ready for exporting
    to .cas animation files which have the opposite sign convention from Milkshape.
    Skeletons with extra bones will have 99 for the hierarchy numbers of the new 
    bones.  Edit the output file to make the desired hierarchy for the new bones.
    """

    showinfo( 'Extract Skeleton Help', infostring )

    return


def InfoexportSkelButton( event = None ):

    infostring = """
       The exportskeleton utility is used to export a skeleton with different pose
    data but the same bone names as a regular skeleton.  For example, to make a
    dwarf or troll animation based on a human animation.  The procedure for this
    goes beyond just running this function as an animation directory must be copied
    and renamed and the file descr_skeleton.txt must be modified to include the new
    animation family and paths changed to point to the new directory.  The utility
    will first display a filechooser for the user to pick a skeleton file.  These 
    are usually obtained by running extractskeleton on a .ms3d file containing the 
    new unit.  Then the utility will display a filechooser for .cas files.  Mouse 
    to the directory that is desired to export to and double-click any file in that
    directory.  The utility will extract the directory name from that file name.
    Finally, the utility will display a data entry dialog for the user to enter scaling
    factors for the x, y, and z components of the animation data.  This allows one
    to adjust the y heights of animations for smaller models like dwarves.

       The converted animation files with the new skeletons and scaled animation data
    will be output in a subdirectory called 'convertedfiles' below the directory that
    the user selected.
    """

    showinfo( 'Export Skeleton Help', infostring )

    return


def InforeorderbonesButton( event = None ):

    infostring = """
       The reorderbones utility is used to process a unit that has weapons and shield
    attachments which have had extra bones added to them.  The reordering moves the
    new bones before the weapon/shield bones so that they may be back converted into
    a mesh file and used in game.  Thus one must use this utility BEFORE extracting the
    skeleton for export and BEFORE using exportextrabonesskeleton to export this skeleton
    to an animation family directory.

       To run the utility click on the reorderbones button.  A filechooser will appear
    with the filter "*.ms3d".  Navigate to the directory containing the file whose bones
    one wishes to reorder, and double-click that file.  If no bones exist beyond the
    weapon/shield bones an information dialog will inform you of this.  Otherwise, an
    ok or cancel dialog will appear asking if the user wishes to proceed.  Click OK to
    procduce a reordered bones .ms3d file.

       The reordered .ms3d file will be output in the same directory as the input .ms3d
    file that the user selected. It will have the same filename except that the string
    "_reordered" will be appended to it.
    """

    showinfo( 'Reorder Bones Help', infostring )

    return


def InfoexportExtraBonesSkelButton( event = None ):

    infostring = """
       The exportextrabonesskeleton utility is used to export a skeleton with 
    extra bones beyond the standard skeleton.  For example, humanoids but with
    tails.  This utility is used after the utility reorderbones has ensured that
    the weapon/shield bones have been moved AFTER any new added bones.             

       To run the utility click on the exportextrabonesskeleton button.  A filechooser
    will appear to select the .skeleton files to export.  After this another filechooser
    will appear.  Navigate to the directory to which the skeleton is desired to be 
    exported to.  Select ANY file in that directory and double-click.  The directory
    name will be extracted from the path.  Finally, a dialog box will appear from which
    the user can enter scaling factors for the animation data to scale the x, y, and z
    data.  Scaling values other than 1.0 are used to account for skeletons with shorter
    or longer limbs than the standard skeleton (think dwarves or trolls).

       The converted animation files with the new skeletons and scaled animation data
    will be output in a subdirectory called 'convertedfiles' below the directory that
    the user selected.
    """

    showinfo( 'Export Extra Bones Skeleton Help', infostring )

    return


def InfoconvertcasButton( event = None ):

    infostring = """
       The convertcastotxt utility is used to read a binary .cas file and write
    out an ASCII .txt file containing all the information in the .cas file but in
    decimal or float format. The utility displays a filechooser where the user  
    can select the .cas file to be converted.

       The format of the converted file is the following.  The header data is 
    printed on one line followed by the filesize minus the header and footer sizes.  
    This is followed by the hierarchy data and then the time ticks data, each on 
    their own line.  The quaternion data is presented bone by bone with a separator 
    line containing the bone index.  The Milkshape equivalent Euler angles in degrees 
    are given on the same line as the quaternion.  The animation data follows next, 
    again separated by bones.  Next comes the skeleton pose data including Scene_Root.
    Finally, the footer data is the last line of the file.

       The file can be edited and then converted back into a binary .cas file by
    the utility converttxttocas.  This might be useful for very minor tweaks or
    trying experiments.  Note that the utility expects the user to change the Euler
    angles and not the quaternions, so only the Eulers are read by converttxttocas.

       The converted animation file has the same name as the .cas file but with 
    the file extension .txt.  The converted file will be in the same directory as
    the selected .cas file.
    """

    showinfo( 'Convert .cas to .txt Help', infostring )

    return


def InfoconverttxtButton( event = None ):

    infostring = """
       The converttxttocas utility is used to read the output of the convertcastotxt
    utility which is an ASCII file containing all the information in a binary .cas
    file but in readable decimal form.  The utility displays a filechooser from which  
    the user can select a .txt file to convert back into the .cas format.  This utility 
    is useful for making small tweaks to an animation without using Milkshape.  Note 
    that the utility expects the user to change the Euler angles and not the quaternions, 
    so only the Euler angles are read by converttxttocas.


       The converted animation file has the same name as the .txt file but with 
    the file extension .cas and with the string '_modified' appended to the name to
    avoid overwriting the original .cas file.  The converted file will be in the same 
    directory as the selected .txt file.
    """

    showinfo( 'Convert .txt to .cas Help', infostring )

    return


def InfosurveycasdirButton( event = None ):

    infostring = """
       The surveycasdirectory utility is used to make ASCII summaries of all
    the .cas files in a directory.  It displays a checkbox dialog from which 
    the user can select which data from the .cas files is to be included in 
    the summary.  It then displays a filechooser from which the user can 
    navigate to a desired directory and double-click on any file in that 
    directory to select that directory.  

       This utility is mainly used for research; it has little practical use for
    modding.

       The summary file is named for the directory with the string '_summary' appended
    to them.  These files are output in the same directory as this utility.
    """

    showinfo( 'Survey cas Directory', infostring )

    return



# -----------------------------------------------------------------------------------
#                            makeCommandMenu()                                      |
# -----------------------------------------------------------------------------------
def makeCommandMenu() :
    CmdBtn                 = Menubutton( mBar, text = 'File' )
    CmdBtn.pack( side = LEFT, padx = '2m' )
    CmdBtn.menu            = Menu( CmdBtn )
    CmdBtn.menu.add_command( label = 'Exit', underline = 0, command = doexit )
    CmdBtn['menu']         = CmdBtn.menu
    return CmdBtn

def doexit():
    exit()
    return

# -----------------------------------------------------------------------------------
#                            makeHelpMenu()                                         |
# -----------------------------------------------------------------------------------
def makeHelpMenu() :
    HelpBtn                = Menubutton( mBar, text = 'Help' )
    HelpBtn.pack( side = LEFT, padx = '2m' )
    HelpBtn.menu           = Menu( HelpBtn )
    HelpBtn.menu.add_command( label = 'About...', underline = 0, command = AboutBox )
    HelpBtn['menu']        = HelpBtn.menu
    return HelpBtn

# -----------------------------------------------------------------------------------
#                            AboutBox()                                             |
# -----------------------------------------------------------------------------------
def AboutBox():

    thisroot               = Tk()
    thisroot.title( 'About Animation Utilities' )
    
    canvas                 = Canvas( thisroot, width = 360, height = 160 )
    img                    = PhotoImage( file = 'ChrysophylaxDives.gif', master = thisroot )
    canvas.create_image( 15,  5, image = img, anchor = NW )
    canvas.create_text( 180, 20, anchor = CENTER, text = 'Version 1.1' )
    canvas.create_text( 180, 35, anchor = CENTER, text = 'Release date: 18 June 2007' )
    canvas.create_text( 180, 50, anchor = CENTER, text = 'Author: KnightErrant' )
    canvas.create_text(  10, 80, anchor = W, text = 'animationutilities are a set of functions for manipulating .cas files and' )
    canvas.create_text(  10,100, anchor = W, text = '.ms3d Milkshape files to create new and modified animations for human' )
    canvas.create_text(  10,120, anchor = W, text = 'units and mounts but not siege engines, yet. Click on the associated' )
    canvas.create_text(  10,140, anchor = W, text = 'Info button to get help for each utility.' )
    canvas.pack()

    root.mainloop()

    return


# ===================================================================================
#                            main()                                                 |
# ===================================================================================

root                       = Tk()
root.title( 'Animation Utilities' )

# Make a menu bar.
mBar                       = Frame( root, borderwidth = 1 )
mBar.pack( fill = X )                      
CmdBtn                     = makeCommandMenu()
HelpBtn                    = makeHelpMenu()
mBar.tk_menuBar( CmdBtn, HelpBtn )


# Make a frame to hold out buttons.
framemain                  = Frame( root, width = 320, height = 440 )
xf                         = Frame( framemain, relief = GROOVE, borderwidth = 4 )
Label(  framemain, text = 'Select Function' ).place( relx = 0.35, rely=0.075, anchor = W )

Button( xf, text = 'Anim Merge'                  , command = animmergeButton            ).grid( row = 1, column = 0, padx = 10, pady = 10, sticky = NSEW )                           
Button( xf, text = 'Anim Extract'                , command = animextractButton          ).grid( row = 2, column = 0, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Extract Skeleton'            , command = extractSkelButton          ).grid( row = 3, column = 0, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Export Skeleton'             , command = exportSkelButton           ).grid( row = 4, column = 0, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Reorder Bones'               , command = reorderbonesButton         ).grid( row = 5, column = 0, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Export Extra Bones Skeleton' , command = exportExtraBonesSkelButton ).grid( row = 6, column = 0, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Convert cas to txt'          , command = convertcasButton           ).grid( row = 7, column = 0, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Convert txt to cas'          , command = converttxtButton           ).grid( row = 8, column = 0, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Survey cas directory'        , command = surveycasdirButton         ).grid( row = 9, column = 0, padx = 10, pady = 10, sticky = NSEW )

Button( xf, text = 'Info'                        , command = InfoanimmergeButton            ).grid( row = 1, column = 1, padx = 10, pady = 10, sticky = NSEW )                           
Button( xf, text = 'Info'                        , command = InfoanimextractButton          ).grid( row = 2, column = 1, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Info'                        , command = InfoextractSkelButton          ).grid( row = 3, column = 1, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Info'                        , command = InfoexportSkelButton           ).grid( row = 4, column = 1, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Info'                        , command = InforeorderbonesButton         ).grid( row = 5, column = 1, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Info'                        , command = InfoexportExtraBonesSkelButton ).grid( row = 6, column = 1, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Info'                        , command = InfoconvertcasButton           ).grid( row = 7, column = 1, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Info'                        , command = InfoconverttxtButton           ).grid( row = 8, column = 1, padx = 10, pady = 10, sticky = NSEW )
Button( xf, text = 'Info'                        , command = InfosurveycasdirButton         ).grid( row = 9, column = 1, padx = 10, pady = 10, sticky = NSEW )

xf.place( relx = 0.15, rely = 0.08, anchor = NW )

framemain.pack()

root.mainloop()
